'''
Function:
    Implementation of NeteaseMusicClient: https://music.163.com/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import json
import copy
import random
import warnings
from .base import BaseMusicClient
from rich.progress import Progress
from ..utils.neteaseutils import EapiCryptoUtils, MUSIC_QUALITIES, DEFAULT_COOKIES
from ..utils import resp2json, seconds2hms, legalizestring, safeextractfromdict, usesearchheaderscookies, cleanlrc, SongInfo
warnings.filterwarnings('ignore')


'''NeteaseMusicClient'''
class NeteaseMusicClient(BaseMusicClient):
    source = 'NeteaseMusicClient'
    def __init__(self, **kwargs):
        super(NeteaseMusicClient, self).__init__(**kwargs)
        self.default_search_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
            'Referer': 'https://music.163.com/',
        }
        self.default_download_headers = {}
        self.default_headers = self.default_search_headers
        self.default_search_cookies = self.default_search_cookies or DEFAULT_COOKIES
        self.default_download_cookies = self.default_download_cookies or DEFAULT_COOKIES
        self._initsession()
    '''_parsewithxiaoqinapi'''
    def _parsewithxiaoqinapi(self, search_result: dict, request_overrides: dict = None):
        # init
        request_overrides, song_id = request_overrides or {}, search_result['id']
        # parse
        for quality in MUSIC_QUALITIES:
            try:
                resp = self.post('https://wyapi-1.toubiec.cn/api/music/url', json={'id': song_id, 'level': quality}, timeout=10, verify=False, **request_overrides)
                resp.raise_for_status()
                download_result = resp2json(resp=resp)
                if ('data' not in download_result) or (not download_result['data']): continue
            except:
                continue
            download_url: str = safeextractfromdict(download_result, ['data', 0, 'url'], '')
            if not download_url: continue
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': {}, 'quality': quality}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['name'], None)),
                singers=legalizestring(', '.join([singer.get('name') for singer in (safeextractfromdict(search_result, ['ar'], []) or []) if isinstance(singer, dict) and singer.get('name')])), album=legalizestring(safeextractfromdict(search_result, ['al', 'name'], None)), 
                ext=download_url.split('?')[0].split('.')[-1], file_size='NULL', identifier=search_result['id'], duration_s=search_result.get('dt', 0) / 1000 if isinstance(search_result.get('dt', 0), (int, float)) else 0,
                duration=seconds2hms(search_result.get('dt', 0) / 1000 if isinstance(search_result.get('dt', 0), (int, float)) else 0), lyric=None, cover_url=safeextractfromdict(search_result, ['al', 'picUrl'], None), 
                download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        # return
        return song_info
    '''_parsewithcggapi'''
    def _parsewithcggapi(self, search_result: dict, request_overrides: dict = None):
        # init
        request_overrides, song_id = request_overrides or {}, search_result['id']
        # safe fetch filesize func
        safe_fetch_filesize_func = lambda meta: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(str(meta.get('size', '0.00MB')).removesuffix('MB').strip()) if isinstance(meta, dict) else 0
        # to seconds func
        to_seconds_func = lambda x: (lambda s: 0 if not s else (lambda p: p[-3]*3600+p[-2]*60+p[-1] if len(p)>=3 else p[0]*60+p[1] if len(p)==2 else p[0] if len(p)==1 else 0)([int(v) for v in re.findall(r'\d+', s.replace('：', ':'))]) if (':' in s or '：' in s) else (lambda h,m,sec,num: (lambda tot: tot if tot>0 else num)(h*3600+m*60+sec))(int(mo.group(1)) if (mo:=re.search(r'(\d+)\s*(?:小时|时|h|hr)', s)) else 0, int(mo.group(1)) if (mo:=re.search(r'(\d+)\s*(?:分钟|分|m|min)', s)) else 0, (int(mo.group(1)) if (mo:=re.search(r'(\d+)\s*(?:秒|s|sec)', s)) else (int(mo.group(1)) if (mo:=re.search(r'(?:分钟|分|m|min)\s*(\d+)\b', s)) else 0)), int(mo.group(0)) if (mo:=re.search(r'\d+', s)) else 0))(str(x).strip().lower())
        # parse
        for quality in MUSIC_QUALITIES:
            for prefix in ['api-v2', 'api-v1', 'api', 'player']:
                try:
                    resp = self.get(url=f'https://{prefix}.cenguigui.cn/api/netease/music_v1.php?id={song_id}&type=json&level={quality}', timeout=10, **request_overrides)
                    resp.raise_for_status()
                    download_result = resp2json(resp=resp)
                    if 'data' not in download_result or (safe_fetch_filesize_func(download_result['data']) < 1): continue
                    break
                except:
                    continue
            download_url: str = safeextractfromdict(download_result, ['data', 'url'], '')
            if not download_url: continue
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': {}, 'quality': quality}, source=self.source, song_name=legalizestring(safeextractfromdict(download_result, ['data', 'name'], None)),
                singers=legalizestring(safeextractfromdict(download_result, ['data', 'artist'], None)), album=legalizestring(safeextractfromdict(download_result, ['data', 'album'], None)), 
                ext=download_url.split('?')[0].split('.')[-1], file_size=safeextractfromdict(download_result, ['data', 'size'], "").removesuffix('MB').strip() + ' MB', identifier=search_result['id'],
                duration_s=to_seconds_func(safeextractfromdict(download_result, ['data', 'duration'], "")), duration=seconds2hms(to_seconds_func(safeextractfromdict(download_result, ['data', 'duration'], ""))),
                lyric=cleanlrc(safeextractfromdict(download_result, ['data', 'lyric'], "")), cover_url=safeextractfromdict(download_result, ['data', 'pic'], ""), download_url=download_url,
                download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        # return
        return song_info
    '''_parsewithbugpkapi'''
    def _parsewithbugpkapi(self, search_result: dict, request_overrides: dict = None):
        # init
        request_overrides, song_id = request_overrides or {}, search_result['id']
        # safe fetch filesize func
        safe_fetch_filesize_func = lambda meta: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(str(meta.get('size', '0.00MB')).removesuffix('MB').strip()) if isinstance(meta, dict) else 0
        # parse
        for quality in MUSIC_QUALITIES:
            try:
                resp = self.get(f'https://api.bugpk.com/api/163_music?ids={song_id}&level={quality}&type=json', timeout=10, **request_overrides)
                resp.raise_for_status()
                download_result = resp2json(resp=resp)
                if 'url' not in download_result or (safe_fetch_filesize_func(download_result) < 1): continue
            except:
                continue
            download_url: str = safeextractfromdict(download_result, ['url'], '')
            if not download_url: continue
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': {}, 'quality': quality}, source=self.source, song_name=legalizestring(safeextractfromdict(download_result, ['name'], None)),
                singers=legalizestring(safeextractfromdict(download_result, ['ar_name'], None)), album=legalizestring(safeextractfromdict(download_result, ['al_name'], None)), ext=download_url.split('?')[0].split('.')[-1],
                file_size=safeextractfromdict(download_result, ['size'], "").removesuffix('MB').strip() + ' MB', identifier=search_result['id'], duration_s=search_result.get('dt', 0) / 1000 if isinstance(search_result.get('dt', 0), (int, float)) else 0,
                duration=seconds2hms(search_result.get('dt', 0) / 1000 if isinstance(search_result.get('dt', 0), (int, float)) else 0), lyric=cleanlrc(safeextractfromdict(download_result, ['lyric'], "")),
                cover_url=safeextractfromdict(download_result, ['pic'], ""), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        # return
        return song_info
    '''_parsewiththirdpartapis'''
    def _parsewiththirdpartapis(self, search_result: dict, request_overrides: dict = None):
        cookies = self.default_cookies or request_overrides.get('cookies')
        if cookies and (cookies != DEFAULT_COOKIES): return SongInfo(source=self.source, raw_data={'quality': MUSIC_QUALITIES[-1]})
        for imp_func in [self._parsewithcggapi, self._parsewithxiaoqinapi, self._parsewithbugpkapi]:
            try:
                song_info_flac = imp_func(search_result, request_overrides)
                if song_info_flac.with_valid_download_url: break
            except:
                song_info_flac = SongInfo(source=self.source, raw_data={'quality': MUSIC_QUALITIES[-1]})
        return song_info_flac
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # search rules
        default_rule = {'s': keyword, 'type': 1, 'limit': 10, 'offset': 0}
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'https://music.163.com/api/cloudsearch/pc'
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            page_rule = copy.deepcopy(default_rule)
            page_rule['limit'] = page_size
            page_rule['offset'] = int(count // page_size) * page_size
            search_urls.append({'url': base_url, 'data': page_rule})
            count += page_size
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: dict = {}, request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        search_meta = copy.deepcopy(search_url)
        search_url = search_meta.pop('url')
        # successful
        try:
            # --search results
            resp = self.post(search_url, **search_meta, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['result']['songs']
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('id' not in search_result): continue
                song_info = SongInfo(source=self.source, raw_data={'quality': MUSIC_QUALITIES[-1]})
                song_info_flac = self._parsewiththirdpartapis(search_result=search_result, request_overrides=request_overrides)
                # ----parse from high to low music quality until successful fetch
                for quality_idx, quality in enumerate(MUSIC_QUALITIES):
                    if quality_idx >= MUSIC_QUALITIES.index(song_info_flac.raw_data['quality']) and song_info_flac.with_valid_download_url: song_info = song_info_flac; break
                    params = {'ids': [search_result['id']], 'level': quality, 'encodeType': 'flac', 'header': json.dumps({"os": "pc", "appver": "", "osver": "", "deviceId": "pyncm!", "requestId": str(random.randrange(20000000, 30000000))})}
                    if quality == 'sky': params['immerseType'] = 'c51'
                    params = EapiCryptoUtils.encryptparams(url='https://interface3.music.163.com/eapi/song/enhance/player/url/v1', payload=params)
                    cookies = {"os": "pc", "appver": "", "osver": "", "deviceId": "pyncm!"}
                    cookies.update(copy.deepcopy(self.default_cookies))
                    try:
                        resp = self.post('https://interface3.music.163.com/eapi/song/enhance/player/url/v1', data={"params": params}, cookies=cookies, **request_overrides)
                        resp.raise_for_status()
                        download_result: dict = resp2json(resp)
                        if ('data' not in download_result) or (not download_result['data']): continue
                    except:
                        continue
                    download_url: str = safeextractfromdict(download_result, ['data', 0, 'url'], '')
                    if not download_url: continue
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}, 'quality': quality}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['name'], None)),
                        singers=legalizestring(', '.join([singer.get('name') for singer in (safeextractfromdict(search_result, ['ar'], []) or []) if isinstance(singer, dict) and singer.get('name')])), album=legalizestring(safeextractfromdict(search_result, ['al', 'name'], None)), 
                        ext=download_url.split('?')[0].split('.')[-1], file_size='NULL', identifier=search_result['id'], duration_s=search_result.get('dt', 0) / 1000 if isinstance(search_result.get('dt', 0), (int, float)) else 0,
                        duration=seconds2hms(search_result.get('dt', 0) / 1000 if isinstance(search_result.get('dt', 0), (int, float)) else 0), lyric=None, cover_url=safeextractfromdict(search_result, ['al', 'picUrl'], None), 
                        download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                    )
                    song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
                    song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                    song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] != 'NULL') else song_info.ext
                    if song_info.with_valid_download_url: break
                if not song_info.with_valid_download_url: song_info = song_info_flac
                if not song_info.with_valid_download_url: continue
                # --lyric results
                data = {'id': search_result['id'], 'cp': 'false', 'tv': '0', 'lv': '0', 'rv': '0', 'kv': '0', 'yv': '0', 'ytv': '0', 'yrv': '0'}
                try:
                    resp = self.post('https://interface3.music.163.com/api/song/lyric', data=data, **request_overrides)
                    resp.raise_for_status()
                    lyric_result: dict = resp2json(resp)
                    lyric = safeextractfromdict(lyric_result, ['lrc', 'lyric'], 'NULL')
                    lyric = 'NULL' if not lyric else cleanlrc(lyric)
                except:
                    lyric_result, lyric = dict(), 'NULL'
                song_info.raw_data['lyric'] = lyric_result
                song_info.lyric = lyric
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of WhisperLRC
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import re
import tempfile
import requests
from typing import Optional, Union, Dict, Any, List


'''settings'''
_LINE_RE = re.compile(r"^\[(\d+),(\d+)\]")
_TOKEN_RE = re.compile(r"<(\d+),(\d+),(\d+)>")
_TIME_RE = re.compile(r"\[(?:(\d{1,2}):)?(\d{1,2}):(\d{2})(?:\.(\d{1,3}))?\]")
cleanlrc = lambda text: "\n".join(line for raw in re.sub(r"\r\n?", "\n", text).split("\n") if (line := raw.strip("\ufeff\u200b\u200c\u200d\u2060\u00a0 \t").strip()) and not re.fullmatch(r"\[(\d{2}:)?\d{2}:\d{2}(?:\.\d{1,3})?\]", line))


'''fractoseconds'''
def fractoseconds(frac: str | None) -> float:
    if not frac: return 0.0
    scale = 10 ** len(frac)
    return int(frac) / scale


'''extractdurationsecondsfromlrc'''
def extractdurationsecondsfromlrc(lrc: str) -> Optional[float]:
    max_t = None
    for h, m, s, frac in _TIME_RE.findall(lrc):
        hh = int(h) if h else 0
        mm = int(m)
        ss = int(s)
        t = hh * 3600 + mm * 60 + ss + fractoseconds(frac)
        max_t = t if (max_t is None or t > max_t) else max_t
    return max_t


'''sectolrcts'''
def sectolrcts(t: Union[str, float, int], centis: int = 2) -> str:
    try: sec = float(t)
    except (TypeError, ValueError): sec = 0.0
    if sec < 0: sec = 0.0
    if centis == 3: return f"{(cs:=int(round(sec*1000)))//60000:02d}:{(cs//1000)%60:02d}.{cs%1000:03d}"
    else: return f"{(cs:=int(round(sec*100)))//6000:02d}:{(cs//100)%60:02d}.{cs%100:02d}"


'''lyricslisttolrc'''
def lyricslisttolrc(items: List[Dict[str, Any]], *, time_key: str = "time", lyric_key: str = "lineLyric", centis: int = 2, offset: float = 0.0, skip_empty: bool = True, dedup_same_time: bool = False, join_sep: str = " / ") -> str:
    norm = []
    for x in items:
        lyric = str(x.get(lyric_key, "")).strip()
        if skip_empty and not lyric: continue
        try: t = float(x.get(time_key, 0.0)) + float(offset)
        except (TypeError, ValueError): t = 0.0
        if t < 0: t = 0.0
        norm.append((t, lyric))
    norm.sort(key=lambda z: z[0])
    if dedup_same_time:
        merged = []
        for t, lyric in norm:
            if merged and abs(merged[-1][0] - t) < 1e-6: merged[-1] = (merged[-1][0], merged[-1][1] + join_sep + lyric)
            else: merged.append((t, lyric))
        norm = merged
    lines = [f"[{sectolrcts(t, centis=centis)}]{lyric}" for t, lyric in norm]
    return "\n".join(lines)


'''TimedLyricsParser'''
class TimedLyricsParser:
    '''parsetimedlyrics'''
    @staticmethod
    def parsetimedlyrics(text: str) -> List[Dict[str, Any]]:
        if not text: return []
        text = text.replace(r"\u003C", "<").replace(r"\u003E", ">")
        lines_out: List[Dict[str, Any]] = []
        for raw_line in text.splitlines():
            raw_line = raw_line.rstrip("\n")
            if not raw_line.strip(): continue
            m = _LINE_RE.match(raw_line.strip())
            if not m: continue
            line_start, line_dur = int(m.group(1)), int(m.group(2))
            line_end, rest, tokens, pieces = line_start + line_dur, raw_line[m.end():], [], []
            matches = list(_TOKEN_RE.finditer(rest))
            for i, tm in enumerate(matches):
                offset, dur, flag, seg_start = int(tm.group(1)), int(tm.group(2)), int(tm.group(3)), tm.end()
                seg_end = matches[i + 1].start() if i + 1 < len(matches) else len(rest)
                token_text = rest[seg_start: seg_end].replace("\r", "")
                if token_text == "": continue
                abs_start, abs_end = line_start + offset, line_start + offset + dur
                tokens.append({"text": token_text, "offset_ms": offset, "duration_ms": dur, "flag": flag, "start_ms": abs_start, "end_ms": abs_end}); pieces.append(token_text)
            lines_out.append({"line_start_ms": line_start, "line_duration_ms": line_dur, "line_end_ms": line_end, "text": "".join(pieces), "tokens": tokens, "raw": rest})
        return lines_out
    '''toplaintext'''
    @staticmethod
    def toplaintext(parsed: List[Dict[str, Any]]) -> str:
        if not parsed: return
        return "\n".join(line["text"] for line in parsed)
    '''tolrclinelevel'''
    @staticmethod
    def tolrclinelevel(parsed: List[Dict[str, Any]], use_centiseconds: bool = True) -> str:
        if not parsed: return
        def fmt(ms: int) -> str:
            mm, ss = ms // 60000, (ms % 60000) // 1000
            if use_centiseconds: xx = (ms % 1000) // 10; return f"{mm:02d}:{ss:02d}.{xx:02d}"
            else: return f"{mm:02d}:{ss:02d}"
        return "\n".join(f"[{fmt(line['line_start_ms'])}]{line['text']}" for line in parsed)


'''WhisperLRC'''
class WhisperLRC:
    def __init__(self, model_size_or_path="small", device="auto", compute_type="int8", cpu_threads=4, num_workers=1, **kwargs):
        try:
            from faster_whisper import WhisperModel
            self.whisper_model = WhisperModel(model_size_or_path, device=device, compute_type=compute_type, cpu_threads=cpu_threads, num_workers=num_workers, **kwargs)
        except:
            self.whisper_model = None
    '''downloadtotmpdir'''
    @staticmethod
    def downloadtotmpdir(url: str, headers: dict = None, timeout: int = 300, cookies: dict = None, request_overrides: dict = None):
        headers, cookies, request_overrides = headers or {}, cookies or {}, request_overrides or {}
        if 'headers' not in request_overrides: request_overrides['headers'] = headers
        if 'timeout' not in request_overrides: request_overrides['timeout'] = timeout
        if 'cookies' not in request_overrides: request_overrides['cookies'] = cookies
        resp = requests.get(url, stream=True, **request_overrides)
        resp.raise_for_status()
        m = re.search(r"\.([a-z0-9]{2,5})(?:\?|$)", url, re.I)
        fd, path = tempfile.mkstemp(suffix="."+(m.group(1).lower() if m else "bin"))
        with os.fdopen(fd, "wb") as fp:
            for ch in resp.iter_content(32768):
                if ch: fp.write(ch)
        return path
    '''timestamp'''
    @staticmethod
    def timestamp(t):
        t = max(0.0, float(t)); mm = int(t//60); ss = t - mm*60
        return f"[{mm:02d}:{ss:05.2f}]"
    '''fromurl'''
    def fromurl(self, url: str, transcribe_overrides: dict = None, headers: dict = None, timeout: int = 300, cookies: dict = None, request_overrides: dict = None):
        assert self.whisper_model is not None, 'faster_whisper should be installed via "pip install "faster_whisper"'
        transcribe_overrides, headers, cookies, request_overrides = transcribe_overrides or {}, headers or {}, cookies or {}, request_overrides or {}
        tmp_file_path = ''
        try:
            tmp_file_path = self.downloadtotmpdir(url, headers=headers, timeout=timeout, cookies=cookies, request_overrides=request_overrides)
            default_transcribe_settings = {'language': None, 'vad_filter': True, 'vad_parameters': dict(min_silence_duration_ms=300), 'chunk_length': 30, 'beam_size': 5}
            default_transcribe_settings.update(transcribe_overrides)
            segs, info = self.whisper_model.transcribe(tmp_file_path, **default_transcribe_settings)
            lrc = "\n".join(f"{self.timestamp(s.start)}{s.text.strip()}" for s in segs)
            result = {"language": info.language, "prob": info.language_probability, "duration": getattr(info, "duration", None), 'lyric': lrc}
            return result
        finally:
            if tmp_file_path and os.path.exists(tmp_file_path): os.remove(tmp_file_path)
    '''fromfilepath'''
    def fromfilepath(self, file_path: str, transcribe_overrides: dict = None):
        assert self.whisper_model is not None, 'faster_whisper should be installed via "pip install "faster_whisper"'
        transcribe_overrides = transcribe_overrides or {}
        default_transcribe_settings = {'language': None, 'vad_filter': True, 'vad_parameters': dict(min_silence_duration_ms=300), 'chunk_length': 30, 'beam_size': 5}
        default_transcribe_settings.update(transcribe_overrides)
        segs, info = self.whisper_model.transcribe(file_path, **default_transcribe_settings)
        lrc = "\n".join(f"{self.timestamp(s.start)}{s.text.strip()}" for s in segs)
        result = {"language": info.language, "prob": info.language_probability, "duration": getattr(info, "duration", None), 'lyric': lrc}
        return result


'''title'''
__title__ = 'musicdl'
'''description'''
__description__ = 'Musicdl: A lightweight music downloader written in pure python'
'''url'''
__url__ = 'https://github.com/CharlesPikachu/musicdl'
'''version'''
__version__ = '2.8.4'
'''author'''
__author__ = 'Zhenchao Jin'
'''email'''
__email__ = 'charlesblwx@gmail.com'
'''license'''
__license__ = 'Apache License 2.0'
'''copyright'''
__copyright__ = 'Copyright 2020-2030 Zhenchao Jin'


'''
Function:
    Implementation of JSInterpreter, refer to https://pytubefix.readthedocs.io/en/latest/index.html
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import json
import math
import datetime
import calendar
import operator
import itertools
import contextlib
import email.utils
import collections
from functools import update_wrapper
from contextlib import suppress as compat_contextlib_suppress


'''constants'''
DATE_FORMATS = (
    '%d %B %Y', '%d %b %Y', '%B %d %Y', '%B %dst %Y', '%B %dnd %Y', '%B %drd %Y', '%B %dth %Y', '%b %d %Y', '%b %dst %Y', '%b %dnd %Y', '%b %drd %Y', '%b %dth %Y',
    '%b %dst %Y %I:%M', '%b %dnd %Y %I:%M', '%b %drd %Y %I:%M', '%b %dth %Y %I:%M', '%Y %m %d', '%Y-%m-%d', '%Y.%m.%d.', '%Y/%m/%d', '%Y/%m/%d %H:%M', '%Y/%m/%d %H:%M:%S',
    '%Y%m%d%H%M', '%Y%m%d%H%M%S', '%Y%m%d', '%Y-%m-%d %H:%M', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%S:%f', '%d.%m.%Y %H:%M', '%d.%m.%Y %H.%M',
    '%Y-%m-%dT%H:%M:%SZ', '%Y-%m-%dT%H:%M:%S.%fZ', '%Y-%m-%dT%H:%M:%S.%f0Z', '%Y-%m-%dT%H:%M:%S', '%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%dT%H:%M', '%b %d %Y at %H:%M',
    '%b %d %Y at %H:%M:%S', '%B %d %Y at %H:%M', '%B %d %Y at %H:%M:%S', '%H:%M %d-%b-%Y',
)
DATE_FORMATS_MONTH_FIRST = list(DATE_FORMATS)
DATE_FORMATS_MONTH_FIRST.extend(['%m-%d-%Y', '%m.%d.%Y', '%m/%d/%Y', '%m/%d/%y', '%m/%d/%Y %H:%M:%S'])
TIMEZONE_NAMES = {'UT': 0, 'UTC': 0, 'GMT': 0, 'Z': 0, 'AST': -4, 'ADT': -3, 'EST': -5, 'EDT': -4, 'CST': -6, 'CDT': -5, 'MST': -7, 'MDT': -6, 'PST': -8, 'PDT': -7}
DATE_FORMATS_DAY_FIRST = list(DATE_FORMATS)
DATE_FORMATS_DAY_FIRST.extend(['%d-%m-%Y', '%d.%m.%Y', '%d.%m.%y', '%d/%m/%Y', '%d/%m/%y', '%d/%m/%Y %H:%M:%S', '%d-%m-%Y %H:%M', '%H:%M %d/%m/%Y'])


'''js2json'''
def js2json(code, vars={}, *, strict=False):
    # constants
    STRING_QUOTES = '\'"`'
    STRING_RE = '|'.join(rf'{q}(?:\\.|[^\\{q}])*{q}' for q in STRING_QUOTES)
    COMMENT_RE = r'/\*(?:(?!\*/).)*?\*/|//[^\n]*\n'
    SKIP_RE = fr'\s*(?:{COMMENT_RE})?\s*'
    INTEGER_TABLE = ((fr'(?s)^(0[xX][0-9a-fA-F]+){SKIP_RE}:?$', 16), (fr'(?s)^(0+[0-7]+){SKIP_RE}:?$', 8))
    # process escape
    def _processescape(match):
        JSON_PASSTHROUGH_ESCAPES = R'"\bfnrtu'
        escape = match.group(1) or match.group(2)
        return (Rf'\{escape}' if escape in JSON_PASSTHROUGH_ESCAPES else R'\u00' if escape == 'x' else '' if escape == '\n' else escape)
    # template substitute
    def _templatesubstitute(match):
        evaluated = js2json(match.group(1), vars, strict=strict)
        if evaluated[0] == '"': return json.loads(evaluated)
        return evaluated
    # fix kv
    def _fixkv(m):
        v: str = m.group(0)
        if v in ('true', 'false', 'null'): return v
        elif v in ('undefined', 'void 0'): return 'null'
        elif v.startswith('/*') or v.startswith('//') or v.startswith('!') or v == ',': return ''
        if v[0] in STRING_QUOTES:
            v = re.sub(r'(?s)\${([^}]+)}', _templatesubstitute, v[1:-1]) if v[0] == '`' else v[1:-1]
            escaped = re.sub(r'(?s)(")|\\(.)', _processescape, v)
            r = f'"{escaped}"'
            return r
        for regex, base in INTEGER_TABLE:
            im = re.match(regex, v)
            if im:
                i = int(im.group(1), base)
                return f'"{i}":' if v.endswith(':') else str(i)
        if v in vars:
            try:
                if not strict: json.loads(vars[v])
            except json.JSONDecodeError:
                return json.dumps(vars[v])
            else:
                return vars[v]
        if not strict: return f'"{v}"'
        raise ValueError(f'Unknown value: {v}')
    # create map
    def _createmap(mobj):
        r = json.dumps(dict(json.loads(js2json(mobj.group(1) or '[]', vars=vars))))
        return r
    # process
    code = re.sub(r'new Map\((\[.*?\])?\)', _createmap, code)
    if not strict:
        code = re.sub(r'new Date\((".+")\)', r'\g<1>', code)
        code = re.sub(r'new \w+\((.*?)\)', lambda m: json.dumps(m.group(0)), code)
        code = re.sub(r'parseInt\([^\d]+(\d+)[^\d]+\)', r'\1', code)
        code = re.sub(r'\(function\([^)]*\)\s*\{[^}]*\}\s*\)\s*\(\s*(["\'][^)]*["\'])\s*\)', r'\1', code)
    # return
    return re.sub(rf'''(?sx)
        {STRING_RE}|
        {COMMENT_RE}|,(?={SKIP_RE}[\]}}])|
        void\s0|(?:(?<![0-9])[eE]|[a-df-zA-DF-Z_$])[.a-zA-Z_$0-9]*|
        \b(?:0[xX][0-9a-fA-F]+|0+[0-7]+)(?:{SKIP_RE}:)?|
        [0-9]+(?={SKIP_RE}:)|
        !+
        ''', _fixkv, code)


'''extracttimezone'''
def extracttimezone(date_str):
    r = r'''(?x)
            ^.{8,}?                                              # >=8 char non-TZ prefix, if present
            (?P<tz>Z|                                            # just the UTC Z, or
                (?:(?<=.\b\d{4}|\b\d{2}:\d\d)|                   # preceded by 4 digits or hh:mm or
                   (?<!.\b[a-zA-Z]{3}|[a-zA-Z]{4}|..\b\d\d))     # not preceded by 3 alpha word or >= 4 alpha or 2 digits
                   [ ]?                                          # optional space
                (?P<sign>\+|-)                                   # +/-
                (?P<hours>[0-9]{2}):?(?P<minutes>[0-9]{2})       # hh[:]mm
            $)
        '''
    m = re.search(r, date_str)
    if not m:
        m = re.search(r'\d{1,2}:\d{1,2}(?:\.\d+)?(?P<tz>\s*[A-Z]+)$', date_str)
        timezone = TIMEZONE_NAMES.get(m and m.group('tz').strip())
        if timezone is not None: date_str = date_str[:-len(m.group('tz'))]
        timezone = datetime.timedelta(hours=timezone or 0)
    else:
        date_str = date_str[:-len(m.group('tz'))]
        if not m.group('sign'): timezone = datetime.timedelta()
        else:
            sign = 1 if m.group('sign') == '+' else -1
            timezone = datetime.timedelta(hours=sign * int(m.group('hours')), minutes=sign * int(m.group('minutes')))
    return timezone, date_str


'''dateformats'''
def dateformats(day_first=True):
    return DATE_FORMATS_DAY_FIRST if day_first else DATE_FORMATS_MONTH_FIRST


'''unifiedtimestamp'''
def unifiedtimestamp(date_str, day_first=True):
    if not isinstance(date_str, str): return None
    date_str = re.sub(r'\s+', ' ', re.sub(r'(?i)[,|]|(mon|tues?|wed(nes)?|thu(rs)?|fri|sat(ur)?)(day)?', '', date_str))
    pm_delta = 12 if re.search(r'(?i)PM', date_str) else 0
    timezone, date_str = extracttimezone(date_str)
    date_str = re.sub(r'(?i)\s*(?:AM|PM)(?:\s+[A-Z]+)?', '', date_str)
    m = re.search(r'\d{1,2}:\d{1,2}(?:\.\d+)?(?P<tz>\s*[A-Z]+)$', date_str)
    if m: date_str = date_str[:-len(m.group('tz'))]
    m = re.search(r'^([0-9]{4,}-[0-9]{1,2}-[0-9]{1,2}T[0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}\.[0-9]{6})[0-9]+$', date_str)
    if m: date_str = m.group(1)
    for expression in dateformats(day_first):
        with contextlib.suppress(ValueError):
            dt = datetime.datetime.strptime(date_str, expression) - timezone + datetime.timedelta(hours=pm_delta)
            return calendar.timegm(dt.timetuple())
    timetuple = email.utils.parsedate_tz(date_str)
    if timetuple: return calendar.timegm(timetuple) + pm_delta * 3600 - timezone.total_seconds()


'''removequotes'''
def removequotes(s):
    if s is None or len(s) < 2:
        return s
    for quote in ('"', "'",):
        if s[0] == quote and s[-1] == quote:
            return s[1:-1]
    return s


'''truncatestring'''
def truncatestring(s, left, right=0):
    assert left > 3 and right >= 0
    if s is None or len(s) <= left + right:
        return s
    return f'{s[:left - 3]}...{s[-right:] if right else ""}'


'''jsbitop'''
def jsbitop(op):
    def zeroise(x):
        if x in (None, JSUndefined):
            return 0
        with contextlib.suppress(TypeError):
            if math.isnan(x): return 0
        return x
    def wrapped(a, b):
        return op(zeroise(a), zeroise(b)) & 0xffffffff
    return wrapped


'''jsarithop'''
def jsarithop(op):
    def wrapped(a, b):
        if JSUndefined in (a, b): return float('nan')
        return op(a or 0, b or 0)
    return wrapped


'''jsdiv'''
def jsdiv(a, b):
    if JSUndefined in (a, b) or not (a or b):
        return float('nan')
    return (a or 0) / b if b else float('inf')


'''jsmod'''
def jsmod(a, b):
    if JSUndefined in (a, b) or not b:
        return float('nan')
    return (a or 0) % b


'''jsexp'''
def jsexp(a, b):
    if not b: return 1
    elif JSUndefined in (a, b): return float('nan')
    return (a or 0) ** b


'''jseqop'''
def jseqop(op):
    def wrapped(a, b):
        if {a, b} <= {None, JSUndefined}: return op(a, a)
        return op(a, b)
    return wrapped


'''jscompop'''
def jscompop(op):
    def wrapped(a, b):
        if JSUndefined in (a, b): return False
        if isinstance(a, str) or isinstance(b, str): return op(str(a or 0), str(b or 0))
        return op(a or 0, b or 0)
    return wrapped


'''jsternary'''
def jsternary(cndn, if_true=True, if_false=False):
    if cndn in (False, None, 0, '', JSUndefined):
        return if_false
    with contextlib.suppress(TypeError):
        if math.isnan(cndn): return if_false
    return if_true


'''jstypeof'''
def jstypeof(expr):
    with compat_contextlib_suppress(TypeError, KeyError):
        return {JSUndefined: 'undefined', float('nan'): 'number', float('inf'): 'number', True: 'boolean', False: 'boolean', None: 'object'}[expr]
    for t, n in (((str, bytes), 'string'), ((int, float, complex), 'number')):
        if isinstance(expr, t): return n
    if callable(expr): return 'function'
    return 'object'


'''wrapsop'''
def wrapsop(op):
    def _updateandrenamewrapper(w):
        f = update_wrapper(w, op)
        f.__name__ = str('JS_') + f.__name__
        return f
    return _updateandrenamewrapper


'''jsunaryop'''
def jsunaryop(op):
    @wrapsop(op)
    def wrapped(_, a):
        return op(a)
    return wrapped


'''extractplayerjsglobalvar'''
def extractplayerjsglobalvar(jscode):
    global_var = re.search(
        r'''(?x)
            (?P<q1>["\'])use\s+strict(?P=q1);\s*
            (?P<code>
                var\s+(?P<name>[a-zA-Z0-9_$]+)\s*=\s*
                (?P<value>
                    (?P<q2>["\'])(?:(?!(?P=q2)).|\\.)+(?P=q2)
                    \.split\((?P<q3>["\'])(?:(?!(?P=q3)).)+(?P=q3)\)
                    |\[\s*(?:(?P<q4>["\'])(?:(?!(?P=q4)).|\\.)*(?P=q4)\s*,?\s*)+\]
                )
            )[;,]
        ''', jscode)
    if global_var: return global_var.group('code'), global_var.group("name"), global_var.group("value")
    else: return None, None, None


'''fixupnfunctioncode'''
def fixupnfunctioncode(argnames, code, full_code):
    global_var, _, _ = extractplayerjsglobalvar(full_code)
    if global_var: code = global_var + '; ' + code
    return argnames, re.sub(rf';\s*if\s*\(\s*typeof\s+[a-zA-Z0-9_$]+\s*===?\s*(["\'])undefined\1\s*\)\s*return\s+{argnames[0]};', ';', code)


'''NODEFAULT'''
class NODEFAULT:
    pass


'''FunctionWithRepr'''
class FunctionWithRepr:
    def __init__(self, func, repr_=None):
        self.func, self.__repr = func, repr_
    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)
    def __repr__(self):
        if self.__repr:
            return self.__repr
        return f'{self.func.__module__}.{self.func.__qualname__}'


'''JSUndefined'''
class JSUndefined:
    pass


'''JSBreak'''
class JSBreak(Exception):
    def __init__(self):
        Exception.__init__(self, 'Invalid break')


'''JSContinue'''
class JSContinue(Exception):
    def __init__(self):
        Exception.__init__(self, 'Invalid continue')


'''JSThrow'''
class JSThrow(Exception):
    def __init__(self, e):
        self.error = e
        Exception.__init__(self, f'Uncaught exception {e}')


'''LocalNameSpace'''
class LocalNameSpace(collections.ChainMap):
    def __setitem__(self, key, value):
        for scope in self.maps:
            if key in scope:
                scope[key] = value
                return
        self.maps[0][key] = value
    def __delitem__(self, key):
        raise NotImplementedError('Deleting is not supported')


'''constants'''
_OPERATORS = {
    '?': None, '??': None, '||': None, '&&': None, '|': jsbitop(operator.or_), '^': jsbitop(operator.xor), '&': jsbitop(operator.and_), '===': operator.is_,
    '!==': operator.is_not, '==': jseqop(operator.eq), '!=': jseqop(operator.ne), '<=': jscompop(operator.le), '>=': jscompop(operator.ge), '<': jscompop(operator.lt),
    '>': jscompop(operator.gt), '>>': jsbitop(operator.rshift), '<<': jsbitop(operator.lshift), '+': jsarithop(operator.add), '-': jsarithop(operator.sub),
    '*': jsarithop(operator.mul), '%': jsmod, '/': jsdiv, '**': jsexp,
}
_UNARY_OPERATORS_X = {'void': jsunaryop(lambda _: JSUndefined), 'typeof': jsunaryop(jstypeof)}
_COMP_OPERATORS = {'===', '!==', '==', '!=', '<=', '>=', '<', '>'}
_ALL_OPERATORS = {**_OPERATORS,  **_UNARY_OPERATORS_X}
_NAME_RE = r'[a-zA-Z_$][\w$]*'
_MATCHING_PARENS = dict(zip(*zip('()', '{}', '[]')))
_QUOTES = '\'"/'
_NESTED_BRACKETS = r'[^[\]]+(?:\[[^[\]]+(?:\[[^\]]+\])?\])?'


'''JSInterpreter'''
class JSInterpreter:
    __named_object_counter = 0
    _RE_FLAGS = {'d': 1024, 'g': 2048, 'i': re.I, 'm': re.M, 's': re.S, 'u': re.U, 'y': 4096}
    def __init__(self, code, objects=None):
        self.code, self._functions = code, {}
        self._objects = {} if objects is None else objects
    '''Exception'''
    class Exception(Exception):
        def __init__(self, msg, expr=None, *args, **kwargs):
            if expr is not None:
                msg = f'{msg.rstrip()} in: {truncatestring(expr, 50, 50)}'
            super().__init__(msg, *args, **kwargs)
    '''_namedobject'''
    def _namedobject(self, namespace, obj):
        self.__named_object_counter += 1
        name = f'__pytubefix_jsinterp_obj{self.__named_object_counter}'
        if callable(obj) and not isinstance(obj, FunctionWithRepr):
            obj = FunctionWithRepr(obj, f'F<{self.__named_object_counter}>')
        namespace[name] = obj
        return name
    '''_regexflags'''
    @classmethod
    def _regexflags(cls, expr):
        flags = 0
        if not expr: return flags, expr
        for idx, ch in enumerate(expr):
            if ch not in cls._RE_FLAGS: break
            flags |= cls._RE_FLAGS[ch]
        return flags, expr[idx + 1:]
    '''_separate'''
    @staticmethod
    def _separate(expr, delim=',', max_split=None):
        OP_CHARS = '+-*/%&|^=<>!,;{}:['
        if not expr: return
        counters = {k: 0 for k in _MATCHING_PARENS.values()}
        start, splits, pos, delim_len = 0, 0, 0, len(delim) - 1
        in_quote, escaping, after_op, in_regex_char_group = None, False, True, False
        for idx, char in enumerate(expr):
            if not in_quote and char in _MATCHING_PARENS:
                counters[_MATCHING_PARENS[char]] += 1
            elif not in_quote and char in counters:
                if counters[char]: counters[char] -= 1
            elif not escaping:
                if char in _QUOTES and in_quote in (char, None):
                    if in_quote or after_op or char != '/': in_quote = None if in_quote and not in_regex_char_group else char
                elif in_quote == '/' and char in '[]':
                    in_regex_char_group = char == '['
            escaping = not escaping and in_quote and char == '\\'
            in_unary_op = (not in_quote and not in_regex_char_group and after_op not in (True, False) and char in '-+')
            after_op = char if (not in_quote and char in OP_CHARS) else (char.isspace() and after_op)
            if char != delim[pos] or any(counters.values()) or in_quote or in_unary_op:
                pos = 0
                continue
            elif pos != delim_len:
                pos += 1
                continue
            yield expr[start: idx - delim_len]
            start, pos = idx + 1, 0
            splits += 1
            if max_split and splits >= max_split: break
        yield expr[start:]
    '''_separateatparen'''
    @classmethod
    def _separateatparen(cls, expr, delim=None):
        if delim is None:
            delim = expr and _MATCHING_PARENS[expr[0]]
        separated = list(cls._separate(expr, delim, 1))
        if len(separated) < 2:
            raise cls.Exception(f'No terminating paren {delim}', expr)
        return separated[0][1:].strip(), separated[1].strip()
    '''_operator'''
    def _operator(self, op, left_val, right_expr, expr, local_vars, allow_recursion):
        if op in ('||', '&&'):
            if (op == '&&') ^ jsternary(left_val): return left_val
        elif op == '??':
            if left_val not in (None, JSUndefined): return left_val
        elif op == '?':
            right_expr = jsternary(left_val, *self._separate(right_expr, ':', 1))
        right_val = self.interpretexpression(right_expr, local_vars, allow_recursion)
        if not _OPERATORS.get(op): return right_val
        try: return _OPERATORS[op](left_val, right_val)
        except Exception as e: raise self.Exception(f'Failed to evaluate {left_val!r} {op} {right_val!r}', expr, cause=e)
    '''_index'''
    def _index(self, obj, idx, allow_undefined=False):
        if idx == 'length': return len(obj)
        try:
            return obj[int(idx)] if isinstance(obj, list) else obj[idx]
        except Exception as e:
            if allow_undefined: return JSUndefined
            raise self.Exception(f'Cannot get index {idx}', repr(obj), cause=e)
    '''_dump'''
    def _dump(self, obj, namespace):
        try: return json.dumps(obj)
        except TypeError: return self._namedobject(namespace, obj)
    '''handleoperators'''
    def handleoperators(self, expr, local_vars, allow_recursion):
        for op in _ALL_OPERATORS:
            separated = list(self._separate(expr, op))
            right_expr = separated.pop()
            while True:
                if op in '?<>*-' and len(separated) > 1 and not separated[-1].strip(): separated.pop()
                elif not (separated and op == '?' and right_expr.startswith('.')): break
                right_expr = f'{op}{right_expr}'
                if op != '-': right_expr = f'{separated.pop()}{op}{right_expr}'
            if not separated: continue
            left_val = self.interpretexpression(op.join(separated), local_vars, allow_recursion)
            return self._operator(op, left_val, right_expr, expr, local_vars, allow_recursion), True
    '''interpretstatement'''
    def interpretstatement(self, stmt, local_vars, allow_recursion=100):
        if allow_recursion < 0: raise self.Exception('Recursion limit reached')
        allow_recursion -= 1
        should_return = False
        sub_statements = list(self._separate(stmt, ';')) or ['']
        expr = stmt = sub_statements.pop().strip()
        for sub_stmt in sub_statements:
            ret, should_return = self.interpretstatement(sub_stmt, local_vars, allow_recursion)
            if should_return: return ret, should_return
        m = re.match(r'(?P<var>(?:var|const|let)\s)|return(?:\s+|(?=["\'])|$)|(?P<throw>throw\s+)', stmt)
        if m:
            expr = stmt[len(m.group(0)):].strip()
            if m.group('throw'): raise JSThrow(self.interpretexpression(expr, local_vars, allow_recursion))
            should_return = not m.group('var')
        if not expr: return None, should_return
        if expr[0] in _QUOTES:
            inner, outer = self._separate(expr, expr[0], 1)
            if expr[0] == '/':
                flags, outer = self._regexflags(outer)
                inner = f'{inner}/{flags}'
            else:
                inner = json.loads(js2json(f'{inner}{expr[0]}', strict=True))
            if not outer: return inner, should_return
            expr = self._namedobject(local_vars, inner) + outer
        if expr.startswith('new '):
            obj = expr[4:]
            if obj.startswith('Date('):
                left, right = self._separateatparen(obj[4:])
                date = unifiedtimestamp(self.interpretexpression(left, local_vars, allow_recursion), False)
                if date is None: raise self.Exception(f'Failed to parse date {left!r}', expr)
                expr = self._dump(int(date * 1000), local_vars) + right
            else:
                raise self.Exception(f'Unsupported object {obj}', expr)
        if expr.startswith('void '):
            left = self.interpretexpression(expr[5:], local_vars, allow_recursion)
            return None, should_return
        for op in _UNARY_OPERATORS_X:
            if not expr.startswith(op): continue
            operand = expr[len(op):]
            if not operand or operand[0] != ' ': continue
            op_result = self.handleoperators(expr, local_vars, allow_recursion)
            if op_result: return op_result[0], should_return
        if expr.startswith('{'):
            inner, outer = self._separateatparen(expr)
            sub_expressions = [list(self._separate(sub_expr.strip(), ':', 1)) for sub_expr in self._separate(inner)]
            if all(len(sub_expr) == 2 for sub_expr in sub_expressions):
                def _dictitem(key, val):
                    val = self.interpretexpression(val, local_vars, allow_recursion)
                    if re.match(_NAME_RE, key): return key, val
                    return self.interpretexpression(key, local_vars, allow_recursion), val
                return dict(_dictitem(k, v) for k, v in sub_expressions), should_return
            inner, should_abort = self.interpretstatement(inner, local_vars, allow_recursion)
            if not outer or should_abort:
                return inner, should_abort or should_return
            else:
                expr = self._dump(inner, local_vars) + outer
        if expr.startswith('('):
            inner, outer = self._separateatparen(expr)
            inner, should_abort = self.interpretstatement(inner, local_vars, allow_recursion)
            if not outer or should_abort: return inner, should_abort or should_return
            else: expr = self._dump(inner, local_vars) + outer
        if expr.startswith('['):
            inner, outer = self._separateatparen(expr)
            name = self._namedobject(local_vars, [self.interpretexpression(item, local_vars, allow_recursion) for item in self._separate(inner)])
            expr = name + outer
        m = re.match(r'''(?x)
                (?P<try>try)\s*\{|
                (?P<if>if)\s*\(|
                (?P<switch>switch)\s*\(|
                (?P<for>for)\s*\(
                ''', expr)
        md = m.groupdict() if m else {}
        if md.get('if'):
            cndn, expr = self._separateatparen(expr[m.end() - 1:])
            if expr.startswith('{'): if_expr, expr = self._separateatparen(expr)
            else: if_expr, expr = self._separateatparen(' %s;' % (expr,), delim=';')
            else_expr = None
            m = re.match(r'else\s*{', expr)
            if m: else_expr, expr = self._separateatparen(expr[m.end() - 1:])
            cndn = jsternary(self.interpretexpression(cndn, local_vars, allow_recursion))
            ret, should_abort = self.interpretstatement(if_expr if cndn else else_expr, local_vars, allow_recursion)
            if should_abort: return ret, True
        if md.get('try'):
            try_expr, expr = self._separateatparen(expr[m.end() - 1:])
            err = None
            try:
                ret, should_abort = self.interpretstatement(try_expr, local_vars, allow_recursion)
                if should_abort: return ret, True
            except Exception as e:
                err = e
            pending = (None, False)
            m = re.match(fr'catch\s*(?P<err>\(\s*{_NAME_RE}\s*\))?\{{', expr)
            if m:
                sub_expr, expr = self._separateatparen(expr[m.end() - 1:])
                if err:
                    catch_vars = {}
                    if m.group('err'): catch_vars[m.group('err')] = err.error if isinstance(err, JSThrow) else err
                    catch_vars = local_vars.new_child(catch_vars)
                    err, pending = None, self.interpretstatement(sub_expr, catch_vars, allow_recursion)
            m = re.match(r'finally\s*\{', expr)
            if m:
                sub_expr, expr = self._separateatparen(expr[m.end() - 1:])
                ret, should_abort = self.interpretstatement(sub_expr, local_vars, allow_recursion)
                if should_abort: return ret, True
            ret, should_abort = pending
            if should_abort: return ret, True
            if err: raise err
        elif md.get('for'):
            constructor, remaining = self._separateatparen(expr[m.end() - 1:])
            if remaining.startswith('{'):
                body, expr = self._separateatparen(remaining)
            else:
                switch_m = re.match(r'switch\s*\(', remaining)  # FIXME
                if switch_m:
                    switch_val, remaining = self._separateatparen(remaining[switch_m.end() - 1:])
                    body, expr = self._separateatparen(remaining, '}')
                    body = 'switch(%s){%s}' % (switch_val, body)
                else:
                    body, expr = remaining, ''
            start, cndn, increment = self._separate(constructor, ';')
            self.interpretexpression(start, local_vars, allow_recursion)
            while True:
                if not jsternary(self.interpretexpression(cndn, local_vars, allow_recursion)): break
                try:
                    ret, should_abort = self.interpretstatement(body, local_vars, allow_recursion)
                    if should_abort: return ret, True
                except JSBreak: break
                except JSContinue: pass
                self.interpretexpression(increment, local_vars, allow_recursion)
        elif md.get('switch'):
            switch_val, remaining = self._separateatparen(expr[m.end() - 1:])
            switch_val = self.interpretexpression(switch_val, local_vars, allow_recursion)
            body, expr = self._separateatparen(remaining, '}')
            items = body.replace('default:', 'case default:').split('case ')[1:]
            for default in (False, True):
                matched = False
                for item in items:
                    case, stmt = (i.strip() for i in self._separate(item, ':', 1))
                    if default: matched = matched or case == 'default'
                    elif not matched: matched = (case != 'default' and switch_val == self.interpretexpression(case, local_vars, allow_recursion))
                    if not matched: continue
                    try:
                        ret, should_abort = self.interpretstatement(stmt, local_vars, allow_recursion)
                        if should_abort: return ret
                    except JSBreak:
                        break
                if matched: break
        if md:
            ret, should_abort = self.interpretstatement(expr, local_vars, allow_recursion)
            return ret, should_abort or should_return
        sub_expressions = list(self._separate(expr))
        if len(sub_expressions) > 1:
            for sub_expr in sub_expressions:
                ret, should_abort = self.interpretstatement(sub_expr, local_vars, allow_recursion)
                if should_abort: return ret, True
            return ret, False
        p =fr'''(?x)
                (?P<out>{_NAME_RE})(?:\[(?P<index>{_NESTED_BRACKETS})\])?\s*
                (?P<op>{"|".join(map(re.escape, set(_OPERATORS) - _COMP_OPERATORS))})?
                =(?!=)(?P<expr>.*)$
            '''
        m = re.match(p, expr)
        if m:
            left_val = local_vars.get(m.group('out'))
            if not m.group('index'):
                local_vars[m.group('out')] = self._operator(m.group('op'), left_val, m.group('expr'), expr, local_vars, allow_recursion)
                return local_vars[m.group('out')], should_return
            elif left_val in (None, JSUndefined):
                raise self.Exception(f'Cannot index undefined variable {m.group("out")}', expr)
            idx = self.interpretexpression(m.group('index'), local_vars, allow_recursion)
            if not isinstance(idx, (int, float)): raise self.Exception(f'List index {idx} must be integer', expr)
            idx = int(idx)
            left_val[idx] = self._operator(m.group('op'), self._index(left_val, idx), m.group('expr'), expr, local_vars, allow_recursion)
            return left_val[idx], should_return
        for m in re.finditer(rf'''(?x)
                (?P<pre_sign>\+\+|--)(?P<var1>{_NAME_RE})|
                (?P<var2>{_NAME_RE})(?P<post_sign>\+\+|--)''', expr):
            var = m.group('var1') or m.group('var2')
            start, end = m.span()
            sign = m.group('pre_sign') or m.group('post_sign')
            ret = local_vars[var]
            local_vars[var] += 1 if sign[0] == '+' else -1
            if m.group('pre_sign'): ret = local_vars[var]
            expr = expr[:start] + self._dump(ret, local_vars) + expr[end:]
        if not expr: return None, should_return
        m = re.match(fr'''(?x)
            (?P<assign>
                (?P<out>{_NAME_RE})(?:\[(?P<index>{_NESTED_BRACKETS})\])?\s*
                (?P<op>{"|".join(map(re.escape, set(_OPERATORS) - _COMP_OPERATORS))})?
                =(?!=)(?P<expr>.*)$
            )|(?P<return>
                (?!if|return|true|false|null|undefined|NaN)(?P<name>{_NAME_RE})$
            )|(?P<attribute>
                (?P<var>{_NAME_RE})(?:
                    (?P<nullish>\?)?\.(?P<member>[^(]+)|
                    \[(?P<member2>{_NESTED_BRACKETS})\]
                )\s*
            )|(?P<indexing>
                (?P<in>{_NAME_RE})\[(?P<idx>.+)\]$
            )|(?P<function>
                (?P<fname>{_NAME_RE})\((?P<args>.*)\)$
            )''', expr)
        if m and m.group('assign'):
            left_val = local_vars.get(m.group('out'))
            if not m.group('index'):
                local_vars[m.group('out')] = self._operator(m.group('op'), left_val, m.group('expr'), expr, local_vars, allow_recursion)
                return local_vars[m.group('out')], should_return
            elif left_val in (None, JSUndefined):
                raise self.Exception(f'Cannot index undefined variable {m.group("out")}', expr)
            idx = self.interpretexpression(m.group('index'), local_vars, allow_recursion)
            if not isinstance(idx, (int, float)):
                raise self.Exception(f'List index {idx} must be integer', expr)
            idx = int(idx)
            left_val[idx] = self._operator(m.group('op'), self._index(left_val, idx), m.group('expr'), expr, local_vars, allow_recursion)
            return left_val[idx], should_return
        elif expr.isdigit():
            return int(expr), should_return
        elif expr == 'break':
            raise JSBreak()
        elif expr == 'continue':
            raise JSContinue()
        elif expr == 'undefined':
            return JSUndefined, should_return
        elif expr == 'NaN':
            return float('NaN'), should_return
        elif m and m.group('return'):
            try:
                return local_vars[m.group('name')], should_return
            except KeyError as e:
                return self.extractglobalvar(e.args[0]), should_return
        with contextlib.suppress(ValueError):
            return json.loads(js2json(expr, strict=True)), should_return
        if m and m.group('indexing'):
            val = local_vars[m.group('in')]
            idx = self.interpretexpression(m.group('idx'), local_vars, allow_recursion)
            return self._index(val, idx), should_return
        op_result = self.handleoperators(expr, local_vars, allow_recursion)
        if op_result: return op_result[0], should_return
        if m and m.group('attribute'):
            variable, member, nullish = m.group('var', 'member', 'nullish')
            if not member: member = self.interpretexpression(m.group('member2'), local_vars, allow_recursion)
            arg_str = expr[m.end():]
            if arg_str.startswith('('): arg_str, remaining = self._separateatparen(arg_str)
            else: arg_str, remaining = None, arg_str
            def assertion(cndn, msg):
                if not cndn: raise self.Exception(f'{member} {msg}', expr)
            def _evalmethod():
                nonlocal member
                types = {'String': str, 'Math': float, 'Array': list}
                obj = local_vars.get(variable, types.get(variable, NODEFAULT))
                if obj is NODEFAULT:
                    if variable not in self._objects:
                        try:
                            self._objects[variable] = self.extractobject(variable, local_vars)
                        except self.Exception:
                            if not nullish:
                                raise
                    obj = self._objects.get(variable, JSUndefined)
                if nullish and obj is JSUndefined: return JSUndefined
                if arg_str is None: return self._index(obj, member, nullish)
                argvals = [self.interpretexpression(v, local_vars, allow_recursion) for v in self._separate(arg_str)]
                if isinstance(obj, type) and member.startswith('prototype.'):
                    new_member, _, func_prototype = member.partition('.')[2].partition('.')
                    assertion(argvals, 'takes one or more arguments')
                    assertion(isinstance(argvals[0], obj), f'needs binding to type {obj}')
                    if func_prototype == 'call':
                        obj, *argvals = argvals
                    elif func_prototype == 'apply':
                        assertion(len(argvals) == 2, 'takes two arguments')
                        obj, argvals = argvals
                        assertion(isinstance(argvals, list), 'second argument needs to be a list')
                    else:
                        raise self.Exception(f'Unsupported Function method {func_prototype}', expr)
                    member = new_member
                if obj == str:
                    if member == 'fromCharCode':
                        assertion(argvals, 'takes one or more arguments')
                        return ''.join(map(chr, argvals))
                    raise self.Exception(f'Unsupported String method {member}', expr)
                elif obj == float:
                    if member == 'pow':
                        assertion(len(argvals) == 2, 'takes two arguments')
                        return argvals[0] ** argvals[1]
                    raise self.Exception(f'Unsupported Math method {member}', expr)
                if member == 'split':
                    assertion(argvals, 'takes one or more arguments')
                    assertion(len(argvals) == 1, 'with limit argument is not implemented')
                    return obj.split(argvals[0]) if argvals[0] else list(obj)
                elif member == 'join':
                    assertion(isinstance(obj, list), 'must be applied on a list')
                    assertion(len(argvals) == 1, 'takes exactly one argument')
                    return argvals[0].join(obj)
                elif member == 'reverse':
                    assertion(not argvals, 'does not take any arguments')
                    obj.reverse()
                    return obj
                elif member == 'slice':
                    assertion(isinstance(obj, (list, str)), 'must be applied on a list or string')
                    assertion(len(argvals) <= 2, 'takes between 0 and 2 arguments')
                    return obj[slice(*argvals, None)]
                elif member == 'splice':
                    assertion(isinstance(obj, list), 'must be applied on a list')
                    assertion(argvals, 'takes one or more arguments')
                    index, howMany = map(int, (argvals + [len(obj)])[:2])
                    if index < 0:
                        index += len(obj)
                    add_items = argvals[2:]
                    res = []
                    for i in range(index, min(index + howMany, len(obj))):
                        res.append(obj.pop(index))
                    for i, item in enumerate(add_items):
                        obj.insert(index + i, item)
                    return res
                elif member == 'unshift':
                    assertion(isinstance(obj, list), 'must be applied on a list')
                    assertion(argvals, 'takes one or more arguments')
                    for item in reversed(argvals): obj.insert(0, item)
                    return obj
                elif member == 'pop':
                    assertion(isinstance(obj, list), 'must be applied on a list')
                    assertion(not argvals, 'does not take any arguments')
                    if not obj: return
                    return obj.pop()
                elif member == 'push':
                    assertion(argvals, 'takes one or more arguments')
                    obj.extend(argvals)
                    return obj
                elif member == 'forEach':
                    assertion(argvals, 'takes one or more arguments')
                    assertion(len(argvals) <= 2, 'takes at-most 2 arguments')
                    f, this = (argvals + [''])[:2]
                    return [f((item, idx, obj), {'this': this}, allow_recursion) for idx, item in enumerate(obj)]
                elif member == 'indexOf':
                    assertion(argvals, 'takes one or more arguments')
                    assertion(len(argvals) <= 2, 'takes at-most 2 arguments')
                    idx, start = (argvals + [0])[:2]
                    try: return obj.index(idx, start)
                    except ValueError: return -1
                elif member == 'charCodeAt':
                    assertion(isinstance(obj, str), 'must be applied on a string')
                    assertion(len(argvals) == 1, 'takes exactly one argument')
                    idx = argvals[0] if isinstance(argvals[0], int) else 0
                    if idx >= len(obj): return None
                    return ord(obj[idx])
                idx = int(member) if isinstance(obj, list) else member
                return obj[idx](argvals, allow_recursion=allow_recursion)
            if remaining:
                ret, should_abort = self.interpretstatement(self._namedobject(local_vars, _evalmethod()) + remaining, local_vars, allow_recursion)
                return ret, should_return or should_abort
            else:
                return _evalmethod(), should_return
        elif m and m.group('function'):
            fname = m.group('fname')
            argvals = [self.interpretexpression(v, local_vars, allow_recursion) for v in self._separate(m.group('args'))]
            if fname in local_vars:
                return local_vars[fname](argvals, allow_recursion=allow_recursion), should_return
            elif fname not in self._functions:
                self._functions[fname] = self.extractfunction(fname)
            return self._functions[fname](argvals, allow_recursion=allow_recursion), should_return
        raise self.Exception(f'Unsupported JS expression {truncatestring(expr, 20, 20) if expr != stmt else ""}', stmt)
    '''interpretexpression'''
    def interpretexpression(self, expr, local_vars, allow_recursion):
        ret, should_return = self.interpretstatement(expr, local_vars, allow_recursion)
        if should_return: raise self.Exception('Cannot return from an expression', expr)
        return ret
    '''extractglobalvar'''
    def extractglobalvar(self, var):
        global_var = re.search(fr'''var\s?{re.escape(var)}=(?P<val>.*?);''', self.code)
        code = global_var.group('val')
        return code
    '''extractobject'''
    def extractobject(self, objname, *global_stack):
        _FUNC_NAME_RE = r'''(?:[a-zA-Z$0-9]+|"[a-zA-Z$0-9]+"|'[a-zA-Z$0-9]+')'''
        obj = {}
        obj_m = re.search(
            r'''(?x)
                (?<![a-zA-Z$0-9.])%s\s*=\s*{\s*
                    (?P<fields>(%s\s*:\s*function\s*\(.*?\)\s*{.*?}(?:,\s*)?)*)
                }\s*;
            ''' % (re.escape(objname), _FUNC_NAME_RE), self.code)
        if not obj_m: raise self.Exception(f'Could not find object {objname}')
        fields = obj_m.group('fields')
        fields_m = re.finditer(
            r'''(?x)
                (?P<key>%s)\s*:\s*function\s*\((?P<args>(?:%s|,)*)\){(?P<code>[^}]+)}
            ''' % (_FUNC_NAME_RE, _NAME_RE), fields)
        for f in fields_m:
            argnames = f.group('args').split(',')
            name = removequotes(f.group('key'))
            obj[name] = FunctionWithRepr(self.buildfunction(argnames, f.group('code'), *global_stack), f'F<{name}>')
        return obj
    '''extractfunctioncode'''
    def extractfunctioncode(self, funcname):
        func_m = re.search(
            r'''(?xs)
                (?:
                    function\s+%(name)s|
                    [{;,]\s*%(name)s\s*=\s*function|
                    (?:var|const|let)\s+%(name)s\s*=\s*function
                )\s*
                \((?P<args>[^)]*)\)\s*
                (?P<code>{.+})''' % {'name': re.escape(funcname)}, self.code)
        if func_m is None: raise self.Exception(f'Could not find JS function "{funcname}"')
        code, _ = self._separateatparen(func_m.group('code'))
        return [x.strip() for x in func_m.group('args').split(',')], code
    '''extractfunction'''
    def extractfunction(self, funcname):
        return FunctionWithRepr(self.extractfunctionfromcode(*fixupnfunctioncode(*self.extractfunctioncode(funcname), self.code)), f'F<{funcname}>')
    '''extractfunctionfromcode'''
    def extractfunctionfromcode(self, argnames, code, *global_stack):
        local_vars = {}
        while True:
            mobj = re.search(r'function\((?P<args>[^)]*)\)\s*{', code)
            if mobj is None: break
            start, body_start = mobj.span()
            body, remaining = self._separateatparen(code[body_start - 1:])
            name = self._namedobject(local_vars, self.extractfunctionfromcode([x.strip() for x in mobj.group('args').split(',')], body, local_vars, *global_stack))
            code = code[:start] + name + remaining
        return self.buildfunction(argnames, code, local_vars, *global_stack)
    '''callfunction'''
    def callfunction(self, funcname, *args):
        return self.extractfunction(funcname)(args)
    '''buildfunction'''
    def buildfunction(self, argnames, code, *global_stack):
        global_stack = list(global_stack) or [{}]
        argnames = tuple(argnames)
        def resf(args, kwargs={}, allow_recursion=100):
            global_stack[0].update(itertools.zip_longest(argnames, args, fillvalue=None))
            global_stack[0].update(kwargs)
            var_stack = LocalNameSpace(*global_stack)
            ret, should_abort = self.interpretstatement(code.replace('\n', ' '), var_stack, allow_recursion - 1)
            if should_abort: return ret
        return resf


'''
Function:
    Implementation of KKWSMusicClient: https://www.kkws.cc/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import copy
from bs4 import BeautifulSoup
from .base import BaseMusicClient
from rich.progress import Progress
from urllib.parse import urlencode, urljoin
from ..utils import legalizestring, usesearchheaderscookies, seconds2hms, searchdictbykey, SongInfo, QuarkParser


'''settings'''
FORMAT_RANK = {
    "DSD": 100, "DSF": 100, "DFF": 100, "WAV": 95, "AIFF": 95, "FLAC": 90, "ALAC": 90, "APE": 88, "WV": 88, "OPUS": 70,
    "AAC": 65, "M4A": 65, "OGG": 60, "VORBIS": 60, "MP3": 50, "WMA": 45,
}


'''KKWSMusicClient'''
class KKWSMusicClient(BaseMusicClient):
    source = 'KKWSMusicClient'
    def __init__(self, **kwargs):
        super(KKWSMusicClient, self).__init__(**kwargs)
        assert self.quark_parser_config.get('cookies'), f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so the songs cannot be downloaded.'
        self.default_search_headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_download_headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # construct search urls
        self.search_size_per_page = min(self.search_size_per_source, 15)
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            search_urls.append(f'https://www.kkws.cc/search.html?key={keyword}&page={int(count // page_size) + 1}')
            count += page_size
        # return
        return search_urls
    '''_parsesearchresultsfromhtml'''
    def _parsesearchresultsfromhtml(self, html_text: str):
        soup = BeautifulSoup(html_text, "lxml")
        search_results, base_url = [], 'https://www.kkws.cc/'
        for li in soup.select("ul.listbox > li"):
            a = li.select_one("h2 a[href]")
            if not a: continue
            href = urljoin(base_url, a["href"])
            title_attr = (a.get("title") or "").strip()
            full_text = a.get_text(" ", strip=True)
            name = title_attr.replace("免费下载", "").strip() if title_attr else full_text
            name = re.sub(r"\s*\[[^\]]+\]\s*", " ", name).strip()
            name = re.sub(r"\s*-\s*\d+(\.\d+)?[KMG]?\s*$", "", name).strip()
            m_fmt = re.search(r"\[([^\]]+)\]", full_text)
            file_format = m_fmt.group(1).strip() if m_fmt else ""
            m_size = re.search(r"-\s*([0-9.]+\s*[KMG]?)", full_text, re.IGNORECASE)
            size = (m_size.group(1).replace(" ", "") if m_size else "").strip()
            ems = li.select("small em")
            share_time, singer = "", ""
            if len(ems) >= 1: share_time = ems[0].get_text(strip=True).replace("分享时间：", "").strip()
            if len(ems) >= 2: singer = ems[-1].get_text(strip=True).replace("演唱：", "").strip()
            m_id = re.search(r"/detail/(\d+)\.html", href)
            item_id = m_id.group(1) if m_id else ""
            search_results.append({
                "id": item_id, "name": name, "format": file_format, "size": size, "share_time": share_time, "singer": singer, "detail_url": href,
            })
        return search_results
    '''_extractlyricsandquark'''
    def _extractlyricsandquark(self, html_text: str):
        soup = BeautifulSoup(html_text, "lxml")
        # lyrics
        textbox, lyrics = soup.select_one("#textbox"), ""
        if textbox:
            for br in textbox.find_all("br"): br.replace_with("\n")
            lyrics = textbox.get_text().strip()
        # quark
        url_map = {}
        for a in soup.select("div.downbox a[onclick]"):
            onclick = a.get("onclick", "").strip()
            if not onclick: continue
            args = re.findall(r"'([^']*)'", onclick)
            if not args: continue
            url, fmt, name = None, None, None
            if onclick.startswith("openModel"):
                if len(args) >= 4:
                    name_fmt, url, fmt = args[1], args[2], args[3] or None
                    if "|" in name_fmt:
                        name, fmt2 = name_fmt.split("|", 1)
                        name = name.strip()
                        fmt = fmt or fmt2.strip()
                    else:
                        name = name_fmt.strip()
            elif onclick.startswith("mbgotourl"):
                if len(args) >= 3:
                    name_fmt, url = args[1], args[2]
                    if "|" in name_fmt:
                        name, fmt = name_fmt.split("|", 1)
                        name, fmt = name.strip(), fmt.strip()
                    else:
                        name = name_fmt.strip()
            if not url or "pan.quark.cn" not in url: continue
            entry = url_map.setdefault(url, {"url": url, "formats": set(), "names": set()})
            for k in FORMAT_RANK.keys(): fmt = k if k.lower() in fmt.lower() else fmt
            if fmt: entry["formats"].add(fmt)
            if name: entry["names"].add(name)
        quark_links = []
        for url, entry in url_map.items():
            quark_links.append({"url": entry["url"], "formats": sorted(entry["formats"]), "names": sorted(entry["names"])})
        quark_links = sorted(quark_links, key=lambda x: FORMAT_RANK.get(x['formats'][0], 0), reverse=True)
        # return
        return {"lyrics": lyrics, "quark_links": quark_links}
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = self._parsesearchresultsfromhtml(resp.text)
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('detail_url' not in search_result):
                    continue
                song_info = SongInfo(source=self.source)
                try:
                    resp = self.get(search_result['detail_url'], **request_overrides)
                    resp.raise_for_status()
                    download_result = self._extractlyricsandquark(resp.text)
                except:
                    continue
                for quark_info in download_result['quark_links']:
                    quark_download_url = quark_info['url']
                    try:
                        download_result['quark_parse_result'], download_url = QuarkParser.parsefromurl(quark_download_url, **self.quark_parser_config)
                        duration = searchdictbykey(download_result['quark_parse_result'], 'duration')
                        duration = [int(float(d)) for d in duration if int(float(d)) > 0]
                        if duration: duration = duration[0]
                        else: duration = 0
                        if not download_url: continue
                        download_url_status = self.quark_audio_link_tester.test(download_url, request_overrides)
                        download_url_status['probe_status'] = self.quark_audio_link_tester.probe(download_url, request_overrides)
                        ext = download_url_status['probe_status']['ext']
                        if ext == 'NULL': ext = 'mp3'
                        song_info = SongInfo(
                            source=self.source, download_url=download_url, download_url_status=download_url_status, default_download_headers=self.quark_default_download_headers,
                            raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, ext=ext, file_size=download_url_status['probe_status']['file_size'],
                            duration_s=duration, duration=seconds2hms(duration), lyric=download_result['lyrics'], album='NULL', identifier=search_result.get('id'),
                            song_name=legalizestring(search_result.get('name', 'NULL'), replace_null_string='NULL'), 
                            singers=legalizestring(search_result.get('singer', 'NULL'), replace_null_string='NULL'), 
                        )
                        if song_info.with_valid_download_url: break
                    except:
                        continue
                if not song_info.with_valid_download_url: continue
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''initialize'''
from .data import SongInfo
from .ip import RandomIPGenerator
from .quarkparser import QuarkParser
from .songinfoutils import SongInfoUtils
from .modulebuilder import BaseModuleBuilder
from .logger import LoggerHandle, colorize, printtable, printfullline, smarttrunctable
from .lyric import WhisperLRC, TimedLyricsParser, extractdurationsecondsfromlrc, lyricslisttolrc, cleanlrc
from .misc import (
    AudioLinkTester, legalizestring, touchdir, seconds2hms, byte2mb, cachecookies, resp2json, isvalidresp, safeextractfromdict, replacefile,
    usedownloadheaderscookies, useparseheaderscookies, usesearchheaderscookies, cookies2dict, cookies2string, estimatedurationwithfilesizebr,
    estimatedurationwithfilelink, searchdictbykey, shortenpathsinsonginfos
)


'''
Function:
    Implementation of KuwoMusicClient: http://www.kuwo.cn/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import copy
import random
from .base import BaseMusicClient
from rich.progress import Progress
from urllib.parse import urlencode, urlparse, parse_qs
from ..utils import legalizestring, resp2json, seconds2hms, usesearchheaderscookies, safeextractfromdict, lyricslisttolrc, cleanlrc, SongInfo


'''KuwoMusicClient'''
class KuwoMusicClient(BaseMusicClient):
    source = 'KuwoMusicClient'
    MUSIC_QUALITIES = [(4000, '4000kflac'), (2000, '2000kflac'), (1000, 'flac'), (320, '320kmp3'), (192, '192kmp3'), (128, '128kmp3')][1:] # 4000kflac is encrypted format
    def __init__(self, **kwargs):
        super(KuwoMusicClient, self).__init__(**kwargs)
        self.default_search_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_parsewithyaohudapi'''
    def _parsewithyaohudapi(self, keyword: str, search_result: dict, request_overrides: dict = None, page_no: int = 1, num: int = 1):
        # init
        if page_no > 1: return SongInfo(source=self.source)
        request_overrides, song_id = request_overrides or {}, str(search_result['MUSICRID']).removeprefix('MUSIC_')
        REQUEST_KEYS = ['zn54xgS3NU0cOKEO0yQ', 'xwT5YzREvIwK8LVZ72n']
        MUSIC_QUALITIES = ["hires", "lossless", "SQ", "exhigh", "standard"]
        # parse
        for quality in MUSIC_QUALITIES:
            try:
                resp = self.get(f"https://api.yaohud.cn/api/music/kuwo?key={random.choice(REQUEST_KEYS)}&msg={keyword}&n={num}&size={quality}", timeout=10, **request_overrides)
                resp.raise_for_status()
                download_result = resp2json(resp=resp)
                if 'data' not in download_result: continue
            except:
                continue
            download_url = safeextractfromdict(download_result, ['data', 'vipmusic', 'url'], '')
            if not download_url: continue
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(download_result, ['data', 'name'], None)),
                singers=legalizestring(safeextractfromdict(download_result, ['data', 'songname'], None)), album=legalizestring(safeextractfromdict(download_result, ['data', 'album'], None)), 
                ext=download_url.split('?')[0].split('.')[-1], file_size='NULL', identifier=song_id, duration_s=safeextractfromdict(search_result, ['DURATION'], 0), duration=seconds2hms(safeextractfromdict(search_result, ['DURATION'], 0)),
                lyric='NULL', cover_url=safeextractfromdict(download_result, ['data', 'picture'], ""), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        return song_info
    '''_parsewithcggapi'''
    def _parsewithcggapi(self, keyword: str, search_result: dict, request_overrides: dict = None, page_no: int = 1, num: int = 1):
        # init
        request_overrides, song_id = request_overrides or {}, str(search_result['MUSICRID']).removeprefix('MUSIC_')
        MUSIC_QUALITIES = ["acc", "wma", "ogg", "standard", "exhigh", "ape", "lossless", "hires", "zp", "hifi", "sur", "jymaster"]
        # safe fetch filesize func
        safe_fetch_filesize_func = lambda meta: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(str(meta.get('size', '0.00MB')).removesuffix('MB').strip()) if isinstance(meta, dict) else 0
        # parse
        for quality in MUSIC_QUALITIES[::-1][3:]:
            try:
                resp = self.get(f"https://kw-api.cenguigui.cn/?id={song_id}&type=song&level={quality}&format=json", timeout=10, **request_overrides)
                resp.raise_for_status()
                download_result = resp2json(resp=resp)
                if 'data' not in download_result or (safe_fetch_filesize_func(download_result['data']) < 1): continue
            except:
                continue
            download_url = safeextractfromdict(download_result, ['data', 'url'], '')
            if not download_url: continue
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(download_result, ['data', 'name'], None)),
                singers=legalizestring(safeextractfromdict(download_result, ['data', 'artist'], None)), album=legalizestring(safeextractfromdict(download_result, ['data', 'album'], None)), 
                ext=download_url.split('?')[0].split('.')[-1], file_size=str(safeextractfromdict(download_result, ['data', 'size'], "")).removesuffix('MB').strip() + ' MB', identifier=song_id,
                duration_s=safeextractfromdict(download_result, ['data', 'duration'], 0), duration=seconds2hms(safeextractfromdict(download_result, ['data', 'duration'], 0)),
                lyric=cleanlrc(safeextractfromdict(download_result, ['data', 'lyric'], "")), cover_url=safeextractfromdict(download_result, ['data', 'pic'], ""), download_url=download_url,
                download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        return song_info
    '''_parsewiththirdpartapis'''
    def _parsewiththirdpartapis(self, keyword: str, search_result: dict, request_overrides: dict = None, page_no: int = 1, num: int = 1):
        if self.default_cookies or request_overrides.get('cookies'): return SongInfo(source=self.source)
        for imp_func in [self._parsewithcggapi, self._parsewithyaohudapi]:
            try:
                song_info_flac = imp_func(keyword, search_result, request_overrides, page_no, num)
                if song_info_flac.with_valid_download_url: break
            except:
                song_info_flac = SongInfo(source=self.source)
        return song_info_flac
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # search rules
        default_rule = {
            "vipver": "1", "client": "kt", "ft": "music", "cluster": "0", "strategy": "2012", "encoding": "utf8", "rformat": "json", "mobi": "1", "issubtitle": "1", "show_copyright_off": "1", "pn": "0", "rn": "10", "all": keyword,
        }
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'http://www.kuwo.cn/search/searchMusicBykeyWord?'
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            page_rule = copy.deepcopy(default_rule)
            page_rule['rn'] = page_size
            page_rule['pn'] = str(int(count // page_size))
            search_urls.append(base_url + urlencode(page_rule))
            count += page_size
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        page_no = int(parse_qs(urlparse(search_url).query, keep_blank_values=True).get('pn')[0]) + 1
        safe_fetch_filesize_func = lambda size: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(size.removesuffix('MB').strip()) if isinstance(size, str) else 0
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['abslist']
            for search_result_idx, search_result in enumerate(search_results):
                # --download results
                if not isinstance(search_result, dict) or ('MUSICRID' not in search_result): continue
                song_info = SongInfo(source=self.source)
                song_info_flac = self._parsewiththirdpartapis(keyword=keyword, search_result=search_result, request_overrides=request_overrides, page_no=page_no, num=search_result_idx)
                for quality in KuwoMusicClient.MUSIC_QUALITIES:
                    if song_info_flac.with_valid_download_url and song_info_flac.ext in ('flac',): song_info = song_info_flac; break
                    try:
                        resp = self.get(f"https://mobi.kuwo.cn/mobi.s?f=web&source=kwplayercar_ar_6.0.0.9_B_jiakong_vh.apk&from=PC&type=convert_url_with_sign&br={quality[1]}&rid={search_result['MUSICRID'].removeprefix('MUSIC_')}&&user=C_APK_guanwang_12609069939969033731", **request_overrides)
                        resp.raise_for_status()
                        download_result = resp2json(resp=resp)
                    except:
                        continue
                    download_url = safeextractfromdict(download_result, ['data', 'url'], '')
                    if not download_url: continue
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['SONGNAME'], None)),
                        singers=legalizestring(safeextractfromdict(search_result, ['ARTIST'], None)), album=legalizestring(safeextractfromdict(search_result, ['ALBUM'], None)), ext=download_url.split('?')[0].split('.')[-1], 
                        file_size='NULL', identifier=search_result['MUSICRID'].removeprefix('MUSIC_'), duration_s=safeextractfromdict(search_result, ['DURATION'], 0), duration=seconds2hms(safeextractfromdict(search_result, ['DURATION'], 0)),
                        lyric='NULL', cover_url=safeextractfromdict(search_result, ['hts_MVPIC'], ""), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                    )
                    song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
                    song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                    song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] != 'NULL') else song_info.ext
                    if song_info_flac.with_valid_download_url and (safe_fetch_filesize_func(song_info.file_size) < safe_fetch_filesize_func(song_info_flac.file_size)): song_info = song_info_flac
                    if song_info.with_valid_download_url: break
                if not song_info.with_valid_download_url: song_info = song_info_flac
                if not song_info.with_valid_download_url: continue
                # --lyric results
                params = {'musicId': search_result['MUSICRID'].removeprefix('MUSIC_'), 'httpsStatus': '1'}
                try:
                    resp = self.get('http://m.kuwo.cn/newh5/singles/songinfoandlrc', params=params, **request_overrides)
                    resp.raise_for_status()
                    lyric_result: dict = resp2json(resp)
                    lyric = cleanlrc(lyricslisttolrc(safeextractfromdict(lyric_result, ['data', 'lrclist'], [])))
                except:
                    lyric_result, lyric = {}, 'NULL'
                song_info.raw_data['lyric'] = lyric_result
                song_info.lyric = lyric
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''initialize'''
from .sources import MusicClientBuilder, BaseMusicClient, BuildMusicClient
from .utils import (
    BaseModuleBuilder, LoggerHandle, AudioLinkTester, WhisperLRC, QuarkParser, SongInfo, SongInfoUtils, RandomIPGenerator, TimedLyricsParser,
    cachecookies, resp2json, isvalidresp, safeextractfromdict, replacefile, printfullline, smarttrunctable, usesearchheaderscookies, byte2mb, 
    usedownloadheaderscookies, useparseheaderscookies, cookies2dict, cookies2string, touchdir, seconds2hms, estimatedurationwithfilesizebr,
    estimatedurationwithfilelink, extractdurationsecondsfromlrc, searchdictbykey, colorize, printtable, legalizestring, lyricslisttolrc, 
    shortenpathsinsonginfos, cleanlrc
)


'''
Function:
    Implementation of GequhaiMusicClient: https://www.gequhai.com/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import base64
import json_repair
from bs4 import BeautifulSoup
from urllib.parse import urljoin
from .base import BaseMusicClient
from rich.progress import Progress
from ..utils import legalizestring, usesearchheaderscookies, resp2json, safeextractfromdict, extractdurationsecondsfromlrc, seconds2hms, searchdictbykey, SongInfo, QuarkParser


'''GequhaiMusicClient'''
class GequhaiMusicClient(BaseMusicClient):
    source = 'GequhaiMusicClient'
    def __init__(self, **kwargs):
        super(GequhaiMusicClient, self).__init__(**kwargs)
        if not self.quark_parser_config.get('cookies'): self.logger_handle.warning(f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so song downloads are restricted and only mp3 files can be downloaded.')
        self.default_search_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # construct search urls
        self.search_size_per_page = min(self.search_size_per_source, 12)
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            if int(count // page_size) + 1 == 1:
                search_urls.append(f'https://www.gequhai.com/s/{keyword}')
            else:
                search_urls.append(f'https://www.gequhai.com/s/{keyword}?page={int(count // page_size) + 1}')
            count += page_size
        # return
        return search_urls
    '''_parsesearchresultsfromhtml'''
    def _parsesearchresultsfromhtml(self, html_text: str):
        soup, base_url = BeautifulSoup(html_text, "html.parser"), "https://www.gequhai.com"
        table = soup.select_one("table#myTables")
        if not table: return []
        results = []
        for tr in table.select("tbody tr"):
            tds = tr.find_all("td")
            if len(tds) < 3: continue
            idx_text = tds[0].get_text(strip=True)
            a = tds[1].find("a")
            title = a.get_text(strip=True) if a else tds[1].get_text(strip=True)
            href: str = a.get("href", "") if a else ""
            play_url = urljoin(base_url, href) if href else ""
            singer = tds[2].get_text(strip=True)
            m = re.search(r"/play/(\d+)", href or "")
            play_id = m.group(1) if m else None
            results.append({
                "index": int(idx_text) if idx_text.isdigit() else idx_text, "title": title, "singer": singer,
                "href": href, "play_url": play_url, "play_id": play_id,
            })
        return results
    '''_decodequarkurl'''
    def _decodequarkurl(self, quark_url: str):
        b64 = quark_url.replace("#", "H")
        return base64.b64decode(b64).decode("utf-8", errors="strict")
    '''_extractappdataandwindowvars'''
    def _extractappdataandwindowvars(self, js_text: str) -> dict:
        out = {}
        m = re.search(r"window\.appData\s*=\s*(\{.*?\})\s*;", js_text, flags=re.S)
        if m:
            app = json_repair.loads(m.group(1))
            out["appData"] = app
            out.update(app)
        for k, v in re.findall(r"window\.(\w+)\s*=\s*'([^']*)'\s*;", js_text): out[k] = v
        for k, v in re.findall(r'window\.(\w+)\s*=\s*"([^"]*)"\s*;', js_text): out[k] = v
        for k, v in re.findall(r"window\.(\w+)\s*=\s*(-?\d+(?:\.\d+)?)\s*;", js_text):
            if k in out: continue
            out[k] = int(v) if re.fullmatch(r"-?\d+", v) else float(v)
        for k, v in re.findall(r"window\.(\w+)\s*=\s*(true|false|null)\s*;", js_text, flags=re.I):
            if k in out: continue
            vv = v.lower()
            out[k] = {"true": True, "false": False, "null": None}[vv]
        if "mp3_title" in out and "mp3_author" in out: out.setdefault("mp3_name", f"{out['mp3_title']}-{out['mp3_author']}")
        if "mp3_extra_url" in out: out["mp3_extra_url_decoded"] = self._decodequarkurl(out["mp3_extra_url"])
        return out
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = self._parsesearchresultsfromhtml(resp.text)
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('play_url' not in search_result):
                    continue
                song_info = SongInfo(source=self.source)
                # ----fetch basic information
                try:
                    resp = self.get(search_result['play_url'], **request_overrides)
                    resp.raise_for_status()
                    download_result = self._extractappdataandwindowvars(resp.text)
                    soup = BeautifulSoup(resp.text, 'lxml')
                except:
                    continue
                # ----parse from quark links
                if self.quark_parser_config.get('cookies'):
                    quark_download_url = download_result.get('mp3_extra_url_decoded', '')
                    try:
                        download_result['quark_parse_result'], download_url = QuarkParser.parsefromdirurl(quark_download_url, **self.quark_parser_config)
                        duration = searchdictbykey(download_result, 'duration')
                        duration = [int(float(d)) for d in duration if int(float(d)) > 0]
                        if duration: duration = duration[0]
                        else: duration = 0
                        if not download_url: raise
                        download_url_status = self.quark_audio_link_tester.test(download_url, request_overrides)
                        download_url_status['probe_status'] = self.quark_audio_link_tester.probe(download_url, request_overrides)
                        ext = download_url_status['probe_status']['ext']
                        if ext == 'NULL': ext = 'mp3'
                        song_info = SongInfo(
                            source=self.source, download_url=download_url, download_url_status=download_url_status, raw_data={'search': search_result, 'download': download_result},
                            default_download_headers=self.quark_default_download_headers, ext=ext, file_size=download_url_status['probe_status']['file_size'],
                            duration_s=duration, duration=seconds2hms(duration),
                        )
                    except:
                        song_info = SongInfo(source=self.source)
                # ----parse from play url
                if not song_info.with_valid_download_url:
                    if 'play_id' not in download_result or not download_result['play_id']: continue
                    song_info = SongInfo(source=self.source)
                    try:
                        headers = {
                            "accept": "application/json, text/javascript, */*; q=0.01", "accept-encoding": "gzip, deflate, br, zstd", "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7",
                            "content-type": "application/x-www-form-urlencoded; charset=UTF-8", "origin": "https://www.gequhai.com", "priority": "u=1, i",
                            "sec-ch-ua": "\"Google Chrome\";v=\"143\", \"Chromium\";v=\"143\", \"Not A(Brand\";v=\"24\"", "x-custom-header": "SecretKey", "x-requested-with": "XMLHttpRequest",
                            "sec-ch-ua-mobile": "?0", "sec-ch-ua-platform": "\"Windows\"", "sec-fetch-dest": "empty", "sec-fetch-mode": "cors", "sec-fetch-site": "same-origin",
                            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Safari/537.36",
                        }
                        resp = self.post('https://www.gequhai.com/api/music', data={'id': download_result['play_id'], 'type': '0'}, headers=headers, **request_overrides)
                        resp.raise_for_status()
                        download_result['api/music'] = resp2json(resp=resp)
                        download_url = safeextractfromdict(download_result['api/music'], ['data', 'url'], '')
                        if not download_url: continue
                        download_url_status = self.audio_link_tester.test(download_url, request_overrides)
                        download_url_status['probe_status'] = self.audio_link_tester.probe(download_url, request_overrides)
                        ext = download_url_status['probe_status']['ext']
                        if ext == 'NULL': download_url.split('?')[0].split('.')[-1] or 'mp3'
                        song_info.update(dict(
                            download_url=download_url, download_url_status=download_url_status, raw_data={'search': search_result, 'download': download_result},
                            ext=ext, file_size=download_url_status['probe_status']['file_size']
                        ))
                    except:
                        continue
                if not song_info.with_valid_download_url: continue
                # ----parse more infos
                try:
                    lrc_div = soup.find("div", id="content-lrc2")
                    lyric, lyric_result = lrc_div.get_text("\n", strip=True), {'lrc_div': str(lrc_div)}
                    duration = seconds2hms(extractdurationsecondsfromlrc(lyric))
                except:
                    lyric, lyric_result, duration = 'NULL', {}, '-:-:-'
                if song_info.duration and song_info.duration != '-:-:-': duration = song_info.duration
                song_info.raw_data['lyric'] = lyric_result
                song_info.update(dict(
                    lyric=lyric, duration=duration, song_name=legalizestring(download_result.get('mp3_title', 'NULL'), replace_null_string='NULL'),
                    singers=legalizestring(download_result.get('mp3_author', 'NULL'), replace_null_string='NULL'), album='NULL',
                    identifier=download_result.get('play_id') or download_result.get('mp3_id'),
                ))
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of SongInfoUtils
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
from pathlib import Path
from .data import SongInfo
from tinytag import TinyTag
from .lyric import WhisperLRC
from .logger import LoggerHandle
from .misc import seconds2hms, byte2mb


'''SongInfoUtils'''
class SongInfoUtils:
    '''fillsongtechinfo'''
    @staticmethod
    def fillsongtechinfo(song_info: SongInfo, logger_handle: LoggerHandle, disable_print: bool) -> SongInfo:
        path = Path(song_info.save_path)
        # correct file size
        size = path.stat().st_size
        song_info.file_size_bytes = size
        song_info.file_size = byte2mb(size=size)
        # tinytag parse
        try:
            tag = TinyTag.get(str(path))
        except Exception as err:
            logger_handle.warning(f'SongInfoUtils.fillsongtechinfo >>> {str(path)} (Err: {err})', disable_print=disable_print)
            return song_info
        if tag.duration:
            song_info.duration_s = int(round(tag.duration))
            song_info.duration = seconds2hms(tag.duration)
        if tag.bitrate:
            song_info.bitrate = int(round(tag.bitrate))
        if tag.samplerate:
            song_info.samplerate = int(tag.samplerate)
        if tag.channels:
            song_info.channels = int(tag.channels)
        if getattr(tag, "codec", None):
            song_info.codec = tag.codec
        elif getattr(tag, "extra", None) and isinstance(tag.extra, dict):
            song_info.codec = tag.extra.get("codec") or tag.extra.get("mime-type")
        # lyric
        if os.environ.get('ENABLE_WHISPERLRC', 'False').lower() == 'true' and ((not song_info.lyric) or (song_info.lyric == 'NULL')):
            lyric_result = WhisperLRC(model_size_or_path='small').fromfilepath(str(path))
            lyric = lyric_result['lyric']
            song_info.lyric = lyric
            song_info.raw_data['lyric'] = lyric_result
        # return
        return song_info


'''
Function:
    Implementation of SodaMusicClient: https://www.douyin.com/qishui/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import copy
import json_repair
from pathlib import Path
from .base import BaseMusicClient
from urllib.parse import urlencode
from rich.progress import Progress
from ..utils.sodautils import AudioDecryptor
from ..utils import legalizestring, byte2mb, resp2json, usesearchheaderscookies, safeextractfromdict, seconds2hms, usedownloadheaderscookies, cleanlrc, SongInfo, TimedLyricsParser


'''SodaMusicClient'''
class SodaMusicClient(BaseMusicClient):
    source = 'SodaMusicClient'
    def __init__(self, **kwargs):
        super(SodaMusicClient, self).__init__(**kwargs)
        self.default_search_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_download'''
    @usedownloadheaderscookies
    def _download(self, song_info: SongInfo, request_overrides: dict = None, downloaded_song_infos: list = [], progress: Progress = None, song_progress_id: int = 0):
        downloaded_song_infos = super()._download(song_info=song_info, request_overrides=request_overrides, downloaded_song_infos=downloaded_song_infos, progress=progress, song_progress_id=song_progress_id)
        with open(song_info.save_path, "rb") as fp: file_data = bytearray(fp.read())
        output_filepath = Path(song_info.save_path)
        output_filepath = output_filepath.parent / f'{output_filepath.stem}.m4a'
        AudioDecryptor.decrypt(file_data=file_data, play_auth=song_info.raw_data['play_auth'], output_filepath=str(output_filepath))
        return downloaded_song_infos
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        self.search_size_per_page = min(self.search_size_per_page, 20)
        # search rules
        default_rule = {
            'aid': '386088', 'app_name': '', 'region': '', 'geo_region': '', 'os_region': '', 'sim_region': '', 'device_id': '', 'cdid': '', 'iid': '', 'version_name': '',
            'version_code': '', 'channel': '', 'build_mode': '', 'network_carrier': '', 'ac': '', 'tz_name': '', 'resolution': '', 'device_platform': '', 'device_type': '',
            'os_version': '', 'fp': '', 'q': keyword, 'cursor': 0, 'search_id': '', 'search_method': 'input', 'debug_params': '', 'from_search_id': '', 'search_scene': '',
        }
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'https://api.qishui.com/luna/pc/search/track?'
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            page_rule = copy.deepcopy(default_rule)
            page_rule['cursor'] = count
            search_urls.append(base_url + urlencode(page_rule))
            count += page_size
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        rank_audio_func = lambda video_list: sorted(video_list, key=lambda x: (x.get('Size'), x.get('Bitrate')), reverse=True)
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['result_groups'][0]['data']
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or not safeextractfromdict(search_result, ['entity', 'track', 'id'], None): continue
                song_info, song_id = SongInfo(source=self.source), search_result['entity']['track']['id']
                try:
                    resp = self.get(f'https://api.qishui.com/luna/pc/track_v2?track_id={song_id}&media_type=track&queue_type=', **request_overrides)
                    resp.raise_for_status()
                    download_result: dict = resp2json(resp)
                    if 'track_player' not in download_result: continue
                    resp = self.get(download_result['track_player']['url_player_info'])
                    resp.raise_for_status()
                    download_result['url_player_info_response'] = resp2json(resp)
                except:
                    continue
                audios = safeextractfromdict(download_result, ['url_player_info_response', 'Result', 'Data', 'PlayInfoList'], [])
                if not audios or not isinstance(audios, list): continue
                audios_sorted: list[dict] = rank_audio_func(audios)
                audios_sorted = [a for a in audios_sorted if (a.get('MainPlayUrl') or a.get('BackupPlayUrl'))]
                if not audios_sorted: continue
                for audio_sorted in audios_sorted:
                    download_url = audio_sorted.get('MainPlayUrl') or audio_sorted.get('BackupPlayUrl')
                    play_auth = safeextractfromdict(audio_sorted, ['PlayAuth'], '')
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}, 'play_auth': play_auth}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['entity', 'track', 'name'], None)),
                        singers=legalizestring(', '.join([singer.get('name') for singer in (safeextractfromdict(search_result, ['entity', 'track', 'artists'], []) or []) if isinstance(singer, dict) and singer.get('name')])),
                        album=legalizestring(safeextractfromdict(search_result, ['entity', 'track', 'album', 'name'], None)), ext=safeextractfromdict(audio_sorted, ['Format'], 'm4a'), file_size_bytes=safeextractfromdict(audio_sorted, ['Size'], 0), 
                        file_size=byte2mb(safeextractfromdict(audio_sorted, ['Size'], 0)), identifier=song_id, duration_s=safeextractfromdict(audio_sorted, ['Duration'], 0), duration=seconds2hms(safeextractfromdict(audio_sorted, ['Duration'], 0)), 
                        lyric=cleanlrc(TimedLyricsParser.tolrclinelevel(TimedLyricsParser.parsetimedlyrics(safeextractfromdict(download_result, ['lyric', 'content'], "")))) or 'NULL', 
                        cover_url=str(safeextractfromdict(search_result, ['entity', 'track', 'album', 'url_cover', 'urls', 0], '')) + str(safeextractfromdict(search_result, ['entity', 'track', 'album', 'url_cover', 'uri'], '')) + '~c5_375x375.jpg', 
                        download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                    )
                    song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
                    song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                    song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] not in ('NULL',)) else song_info.ext
                    if song_info.with_valid_download_url: break
                if not song_info.with_valid_download_url: continue
                # --lyric results
                try:
                    resp = self.get(f'https://music.douyin.com/qishui/share/track?track_id={song_id}', **request_overrides)
                    resp.raise_for_status()
                    lyric_result = json_repair.loads(re.search(r'_ROUTER_DATA\s*=\s*({[\s\S]*?});', resp.text).group(1).strip())
                    sentences, lrc_list = lyric_result['loaderData']['track_page']['audioWithLyricsOption']['lyrics']['sentences'], []
                    for sentence in sentences:
                        if not isinstance(sentence, dict): continue
                        start_ms = sentence.get('startMs', 0)
                        sentence_text = "".join([w.get('text', '') for w in sentence.get('words', []) if isinstance(w, dict)])
                        minutes, seconds, m_seconds = start_ms // 60000, (start_ms % 60000) // 1000, start_ms % 1000
                        time_tag = f"[{minutes:02d}:{seconds:02d}.{m_seconds:03d}]"
                        lrc_list.append(f"{time_tag}{sentence_text}")
                    lyric = cleanlrc("\n".join(lrc_list)) or 'NULL'
                except:
                    lyric_result, lyric = {}, 'NULL'
                # --update song_info
                song_info.raw_data['lyric'] = lyric_result
                song_info.lyric = lyric
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        return song_infos


'''
Function:
    Implementation of RandomIPGenerator
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import random
import requests
import ipaddress
from bisect import bisect
from typing import List, Optional, Sequence


'''RandomIPGenerator'''
class RandomIPGenerator:
    def __init__(self, default_ipv4_prefixes: Optional[Sequence[str]] = None, default_ipv6_prefixes: Optional[Sequence[str]] = None, max_attempts: int = 10000):
        self.max_attempts = max_attempts
        self.default_ipv4_prefixes: List[str] = list(default_ipv4_prefixes or [])
        self.default_ipv6_prefixes: List[str] = list(default_ipv6_prefixes or [])
    '''ipv4'''
    def ipv4(self, prefix: Optional[str] = None) -> str:
        if prefix is None and self.default_ipv4_prefixes:
            prefix = random.choice(self.default_ipv4_prefixes)
        if prefix is not None:
            return self._randomipv4inprefix(prefix)
        else:
            return self._randomglobalipv4()
    '''ipv6'''
    def ipv6(self, prefix: Optional[str] = None) -> str:
        if prefix is None and self.default_ipv6_prefixes:
            prefix = random.choice(self.default_ipv6_prefixes)
        if prefix is not None:
            return self._randomipv6inprefix(prefix)
        else:
            return self._randomglobalipv6()
    '''randomipv4scn'''
    def randomipv4scn(self, num_samples: int = 1) -> List[str]:
        def _buildsampler(blocks):
            cum, s = [], 0
            for _, c in blocks: s += c; cum.append(s)
            total = s
            def sample(n=10):
                out = []
                for _ in range(n):
                    r = random.randrange(total)
                    idx = bisect(cum, r)
                    base, count = blocks[idx]
                    ip = ipaddress.IPv4Address(base + random.randrange(count))
                    out.append(str(ip))
                return out
            return sample
        blocks = self._loadcnipv4blocks()
        sampler = _buildsampler(blocks)
        return sampler(num_samples)
    '''addrandomipv4toheaders'''
    def addrandomipv4toheaders(self, headers: dict = None, prefix: Optional[str] = None) -> dict:
        assert isinstance(headers, dict), f'input "headers" should be "dict", but get {type(headers)}'
        random_ip = self.ipv4(prefix=prefix)
        headers.update({
            "X-Forwarded-For": random_ip, "X-Real-IP": random_ip, "Forwarded": f"for={random_ip};proto=https",
        })
        return headers
    '''_loadcnipv4blocks'''
    def _loadcnipv4blocks(self):
        text = requests.get("https://ftp.apnic.net/stats/apnic/delegated-apnic-extended-latest", timeout=30).text.splitlines()
        blocks = []
        for line in text:
            if not line or line.startswith("#"): continue
            parts = line.strip().split("|")
            if len(parts) < 7: continue
            _, cc, rtype, start, value, _, status = parts[:7]
            if cc != "CN" or rtype != "ipv4": continue
            if status not in ("allocated", "assigned"): continue
            base = int(ipaddress.IPv4Address(start))
            count = int(value)
            if count > 0: blocks.append((base, count))
        if not blocks: raise RuntimeError("No CN IPv4 blocks found. Check APNIC file format/URL.")
        return blocks
    '''_randomipv4inprefix'''
    def _randomipv4inprefix(self, prefix: str) -> str:
        net = ipaddress.IPv4Network(prefix, strict=False)
        if net.prefixlen <= 30:
            network_int = int(net.network_address)
            broadcast_int = int(net.broadcast_address)
            if broadcast_int - network_int <= 2: candidate_int = random.randint(network_int, broadcast_int)
            else: candidate_int = random.randint(network_int + 1, broadcast_int - 1)
        else:
            offset = random.randrange(net.num_addresses)
            candidate_int = int(net.network_address) + offset
        addr = ipaddress.IPv4Address(candidate_int)
        return str(addr)
    '''_randomglobalipv4'''
    def _randomglobalipv4(self) -> str:
        attempts = 0
        while attempts < self.max_attempts:
            attempts += 1
            candidate_int = random.getrandbits(32)
            addr = ipaddress.IPv4Address(candidate_int)
            if addr.is_global: return str(addr)
        return str(addr)
    '''_randomipv6inprefix'''
    def _randomipv6inprefix(self, prefix: str) -> str:
        net = ipaddress.IPv6Network(prefix, strict=False)
        host_bits = 128 - net.prefixlen
        rand_host = random.getrandbits(host_bits)
        addr_int = int(net.network_address) + rand_host
        addr = ipaddress.IPv6Address(addr_int)
        return str(addr)
    '''_randomglobalipv6'''
    def _randomglobalipv6(self) -> str:
        attempts = 0
        while attempts < self.max_attempts:
            attempts += 1
            candidate_int = random.getrandbits(128)
            addr = ipaddress.IPv6Address(candidate_int)
            if addr.is_global: return str(addr)
        return str(addr)


'''
Function:
    Implementation of MyFreeMP3MusicClient: https://www.myfreemp3.com.cn/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import copy
from urllib.parse import urlparse
from rich.progress import Progress
from ..sources import BaseMusicClient
from ..utils import legalizestring, resp2json, usesearchheaderscookies, seconds2hms, extractdurationsecondsfromlrc, searchdictbykey, SongInfo, QuarkParser


'''MyFreeMP3MusicClient'''
class MyFreeMP3MusicClient(BaseMusicClient):
    source = 'MyFreeMP3MusicClient'
    def __init__(self, **kwargs):
        super(MyFreeMP3MusicClient, self).__init__(**kwargs)
        if not self.quark_parser_config.get('cookies'): self.logger_handle.warning(f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so only "netease" source can be leveraged.')
        self.allowed_music_sources = ['kuake', 'netease'] if self.quark_parser_config.get('cookies') else ['netease']
        self.default_search_headers = {
            "accept": "*/*", "accept-encoding": "gzip, deflate, br, zstd", "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8", "origin": "https://www.myfreemp3.com.cn",
            "priority": "u=1, i", "sec-ch-ua": "\"Google Chrome\";v=\"143\", \"Chromium\";v=\"143\", \"Not A(Brand\";v=\"24\"",
            "sec-ch-ua-mobile": "?0", "sec-ch-ua-platform": "\"Windows\"", "sec-fetch-dest": "empty", "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin", "x-requested-with": "XMLHttpRequest",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Safari/537.36",
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        allowed_music_sources = copy.deepcopy(self.allowed_music_sources)
        self.search_size_per_page = min(10, self.search_size_per_page)
        # search rules
        default_rule = {'type': 'netease', 'filter': 'name', 'page': '1', 'input': keyword}
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'https://www.myfreemp3.com.cn/'
        search_urls, page_size = [], self.search_size_per_page
        for source in allowed_music_sources:
            source_default_rule = copy.deepcopy(default_rule)
            source_default_rule['type'], count = source, 0
            while self.search_size_per_source > count:
                page_rule = copy.deepcopy(source_default_rule)
                page_rule['page'] = str(int(count // page_size) + 1)
                search_urls.append({'url': base_url, 'data': page_rule, 'source': source})
                count += page_size
        # return
        return search_urls
    '''_parseneteasesearchresult'''
    def _parseneteasesearchresult(self, search_result: dict, request_overrides: dict = None):
        request_overrides = request_overrides or {}
        if (not isinstance(search_result, dict)) or ('id' not in search_result): return SongInfo(source=self.source)
        download_url =  f'http://music.163.com/song/media/outer/url?id={search_result["id"]}.mp3'
        lyric: str = search_result.get('lrc', '')
        lyric = lyric.removeprefix('data:text/plain,')
        duration = extractdurationsecondsfromlrc(lyric)
        song_info = SongInfo(
            source=self.source, download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            ext='mp3', file_size='NULL', duration_s=duration, duration=seconds2hms(duration), root_source='netease', identifier=search_result['id'],
            raw_data={'search': search_result, 'download': {}, 'lyric': {}}, lyric=lyric, cover_url=search_result.get('pic'), album='NULL',
            song_name=legalizestring(search_result.get('title', 'NULL'), replace_null_string='NULL'),
            singers=legalizestring(search_result.get('author', 'NULL'), replace_null_string='NULL'),
        )
        song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
        ext, file_size = song_info.download_url_status['probe_status']['ext'], song_info.download_url_status['probe_status']['file_size']
        if file_size and file_size != 'NULL': song_info.file_size = file_size
        if ext and ext != 'NULL': song_info.ext = ext
        return song_info
    '''_parsequarksearchresult'''
    def _parsequarksearchresult(self, search_result: dict, request_overrides: dict = None):
        request_overrides = request_overrides or {}
        if (not isinstance(search_result, dict)) or ('url_kk' not in search_result): return SongInfo(source=self.source)
        search_result['id'] = urlparse(search_result['url_kk']).path.strip('/').split('/')[-1]
        quark_download_url = search_result['url_kk']
        download_result, download_url = QuarkParser.parsefromurl(quark_download_url, **self.quark_parser_config)
        if not download_url: return SongInfo(source=self.source)
        duration = searchdictbykey(download_result, 'duration')
        duration = [int(float(d)) for d in duration if int(float(d)) > 0]
        if duration: duration = duration[0]
        else: duration = 0
        song, artist = (lambda s: (m.group(2).strip(), m.group(1).strip()) if (m:=re.search(r'^\s*(.*?)\s*[-–—－]\s*(.*?)(?:\.[A-Za-z0-9]{1,5})?\s*(?:\s*[-–—－]\s*.*)?$', s.strip())) else (re.sub(r'\.[^.]+$', '', s.strip()).strip(), ""))(search_result.get('title'))
        song_info = SongInfo(
            source=self.source, download_url=download_url, download_url_status=self.quark_audio_link_tester.test(download_url, request_overrides),
            ext='mp3', file_size='NULL', duration_s=duration, duration=seconds2hms(duration), root_source='quark', identifier=search_result['id'],
            raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, lyric='NULL', album='NULL', default_download_headers=self.quark_default_download_headers,
            song_name=legalizestring(song, replace_null_string='NULL'), singers=legalizestring(artist, replace_null_string='NULL')
        )
        song_info.download_url_status['probe_status'] = self.quark_audio_link_tester.probe(song_info.download_url, request_overrides)
        ext, file_size = song_info.download_url_status['probe_status']['ext'], song_info.download_url_status['probe_status']['file_size']
        if file_size and file_size != 'NULL': song_info.file_size = file_size
        if ext and ext != 'NULL': song_info.ext = ext
        return song_info
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: dict = None, request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        search_meta = copy.deepcopy(search_url)
        search_url, source = search_meta.pop('url'), search_meta.pop('source')
        # successful
        try:
            # --search results
            resp = self.post(search_url, **search_meta, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['data']['list']
            for search_result in search_results:
                # --download results
                if source in ['netease']:
                    try:
                        song_info = self._parseneteasesearchresult(search_result, request_overrides)
                    except:
                        song_info = SongInfo(source=self.source)
                else:
                    assert source in ['kuake']
                    try:
                        song_info = self._parsequarksearchresult(search_result, request_overrides)
                    except:
                        song_info = SongInfo(source=self.source)
                if not song_info.with_valid_download_url: continue
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of YinyuedaoMusicClient: https://1mp3.top/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import json_repair
from bs4 import BeautifulSoup
from .base import BaseMusicClient
from rich.progress import Progress
from ..utils import legalizestring, usesearchheaderscookies, safeextractfromdict, seconds2hms, searchdictbykey, SongInfo, QuarkParser


'''settings'''
FORMAT_RANK = {
    "DSD": 100, "DSF": 100, "DFF": 100, "WAV": 95, "AIFF": 95, "FLAC": 90, "ALAC": 90, "APE": 88, "WV": 88, "OPUS": 70,
    "AAC": 65, "M4A": 65, "OGG": 60, "VORBIS": 60, "MP3": 50, "WMA": 45,
}


'''YinyuedaoMusicClient'''
class YinyuedaoMusicClient(BaseMusicClient):
    source = 'YinyuedaoMusicClient'
    def __init__(self, **kwargs):
        super(YinyuedaoMusicClient, self).__init__(**kwargs)
        if not self.quark_parser_config.get('cookies'): self.logger_handle.warning(f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so song downloads are restricted and only mp3 files can be downloaded.')
        self.default_search_headers = {
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7",
            "priority": "u=0, i",
            "referer": "https://1mp3.top/",
            "sec-ch-ua": "\"Chromium\";v=\"142\", \"Google Chrome\";v=\"142\", \"Not_A Brand\";v=\"99\"",
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": "\"Windows\"",
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "same-origin",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # construct search urls
        search_urls = [f'https://1mp3.top/?key={keyword}&search=1']
        self.search_size_per_page = self.search_size_per_source
        # return
        return search_urls
    '''_parsesearchresultsfromhtml'''
    def _parsesearchresultsfromhtml(self, html_text: str):
        soup = BeautifulSoup(html_text, "lxml")
        search_results = []
        for li in soup.select("#musicList > li"):
            music_id_attr = li.get("data-music-id")
            music_title_attr = li.get("data-music-title")
            music_singer_attr = li.get("data-music-singer")
            music_cover_attr = li.get("data-music-cover")
            a_download = li.select_one("a.download-btn")
            music_json_str = a_download.get("data-music")
            music_data = json_repair.loads(music_json_str)
            search_results.append({
                "id_attr": music_id_attr, "title_attr": music_title_attr, "singer_attr": music_singer_attr, "cover_attr": music_cover_attr,
                "id": music_data.get("id"), "title": music_data.get("title"), "singer": music_data.get("singer"), "picurl": music_data.get("picurl"),
                "create_time": music_data.get("create_time"), "mtype": music_data.get("mtype"), "downlist": music_data.get("downlist", []),
                "ktmdownlist": music_data.get("ktmdownlist", []),
            })
        return search_results
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        def _sortbyformat(items):
            def _score(x):
                fmt = (x.get("format") or "").upper()
                return FORMAT_RANK.get(fmt, 0)
            return sorted(items, key=_score, reverse=True)
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = self._parsesearchresultsfromhtml(resp.text)
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('id' not in search_result):
                    continue
                song_info = SongInfo(source=self.source)
                # ----parse from quark links
                if self.quark_parser_config.get('cookies'):
                    quark_download_urls = [*search_result.get('downlist', []), *search_result.get('ktmdownlist', [])]
                    quark_download_urls = _sortbyformat(quark_download_urls)
                    for quark_download_url in quark_download_urls:
                        song_fmt = safeextractfromdict(quark_download_url, ['format'], '')
                        if not song_fmt: continue
                        song_info = SongInfo(source=self.source)
                        try:
                            quark_wav_download_url = quark_download_url['url']
                            download_result, download_url = QuarkParser.parsefromurl(quark_wav_download_url, **self.quark_parser_config)
                            duration = searchdictbykey(download_result, 'duration')
                            duration = [int(float(d)) for d in duration if int(float(d)) > 0]
                            if duration: duration = duration[0]
                            else: duration = 0
                            if not download_url: continue
                            download_url_status = self.quark_audio_link_tester.test(download_url, request_overrides)
                            download_url_status['probe_status'] = self.quark_audio_link_tester.probe(download_url, request_overrides)
                            ext = download_url_status['probe_status']['ext']
                            if ext == 'NULL': ext = 'mp3'
                            song_info.update(dict(
                                download_url=download_url, download_url_status=download_url_status, raw_data={'search': search_result, 'download': download_result},
                                default_download_headers=self.quark_default_download_headers, ext=ext, file_size=download_url_status['probe_status']['file_size'],
                                duration_s=duration, duration=seconds2hms(duration),
                            ))
                            if song_info.with_valid_download_url: break
                        except:
                            continue
                # ----parse from play url
                if not song_info.with_valid_download_url:
                    song_info = SongInfo(source=self.source)
                    try:
                        resp = self.get(f'https://1mp3.top/include/geturl.php?id={search_result["id"]}', **request_overrides)
                        resp.raise_for_status()
                        download_url = resp.text.strip()
                        download_url_status = self.audio_link_tester.test(download_url, request_overrides)
                        download_url_status['probe_status'] = self.audio_link_tester.probe(download_url, request_overrides)
                        ext = download_url_status['probe_status']['ext']
                        if ext == 'NULL': download_url.split('?')[0].split('.')[-1] or 'mp3'
                        song_info.update(dict(
                            download_url=download_url, download_url_status=download_url_status, raw_data={'search': search_result, 'download': {}},
                            ext=ext, file_size=download_url_status['probe_status']['file_size']
                        ))
                    except:
                        continue
                if not song_info.with_valid_download_url: continue
                # ----parse more infos
                lyric_result, lyric = dict(), 'NULL'
                song_info.raw_data['lyric'] = lyric_result
                song_info.update(dict(
                    lyric=lyric, song_name=legalizestring(search_result.get('title', 'NULL'), replace_null_string='NULL'), 
                    singers=legalizestring(search_result.get('singer', 'NULL'), replace_null_string='NULL'), album='NULL', identifier=search_result['id'],
                ))
                if not song_info.duration or song_info.duration == 'NULL': song_info.duration = '-:-:-'
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of Setup
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
GitHub:
    https://github.com/CharlesPikachu/musicdl
'''
import musicdl
from setuptools import setup, find_packages


'''readme'''
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()


'''setup'''
setup(
    name=musicdl.__title__,
    version=musicdl.__version__,
    description=musicdl.__description__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent'
    ],
    author=musicdl.__author__,
    url=musicdl.__url__,
    author_email=musicdl.__email__,
    license=musicdl.__license__,
    include_package_data=True,
    packages=find_packages(),
    package_data={"musicdl": ["modules/js/youtube/*.js"]},
    entry_points={'console_scripts': ['musicdl = musicdl.musicdl:MusicClientCMD']},
    install_requires=[lab.strip('\n') for lab in list(open('requirements.txt', 'r').readlines())],
    zip_safe=True,
)


'''
Function:
    Implementation of XimalayaMusicClient: https://www.ximalaya.com/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import time
import copy
import base64
import binascii
from Crypto.Cipher import AES
from .base import BaseMusicClient
from rich.progress import Progress
from urllib.parse import urlencode, urlparse
from ..utils import byte2mb, resp2json, seconds2hms, legalizestring, safeextractfromdict, usesearchheaderscookies, SongInfo


'''XimalayaMusicClient'''
class XimalayaMusicClient(BaseMusicClient):
    source = 'XimalayaMusicClient'
    def __init__(self, **kwargs):
        super(XimalayaMusicClient, self).__init__(**kwargs)
        self.default_search_headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_download_headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_decrypturl'''
    def _decrypturl(self, ciphertext: str):
        if not ciphertext: return ciphertext
        key = binascii.unhexlify("aaad3e4fd540b0f79dca95606e72bf93")
        ciphertext = base64.urlsafe_b64decode(ciphertext + "=" * (4 - len(ciphertext) % 4))
        cipher = AES.new(key, AES.MODE_ECB)
        plaintext = cipher.decrypt(ciphertext)
        plaintext = re.sub(r"[^\x20-\x7E]", "", plaintext.decode("utf-8"))
        return plaintext
    '''_validategdstudio'''
    def _validategdstudio(self, request_overrides: dict = None):
        request_overrides = request_overrides or {}
        try:
            resp = self.get('https://music-api.gdstudio.xyz/api.php?types=search&source=ximalaya&name=%E4%B8%89%E5%9B%BD&count=1&pages=1', timeout=10, **request_overrides)
            resp.raise_for_status()
            result = resp2json(resp=resp)
            assert isinstance(result, list) and (len(result) == 1)
            return True
        except:
            return False
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # if with cookies, use official apis
        if self.default_search_cookies:
            # --search rules
            default_rule = {
                'kw': keyword, 'page': 1, 'spellchecker': 'true', 'condition': 'relation', 'rows': self.search_size_per_page, 'device': 'iPhone', 
                'core': 'track', 'fq': '', 'paidFilter': 'false', 
            }
            default_rule.update(rule)
            # --construct search urls based on search rules
            base_url = 'https://www.ximalaya.com/revision/search/main?'
            search_urls, page_size, count = [], self.search_size_per_page, 0
            while self.search_size_per_source > count:
                page_rule = copy.deepcopy(default_rule)
                page_rule['rows'] = page_size
                page_rule['page'] = int(count // page_size) + 1
                search_urls.append(base_url + urlencode(page_rule))
                count += page_size
        # if without cookies, use third part apis
        else:
            # --search rules
            default_rule = {'types': 'search', 'source': 'ximalaya', 'name': keyword, 'count': self.search_size_per_page, 'pages': '1'}
            default_rule.update(rule)
            # --construct search urls based on search rules
            base_url = 'https://music-api.gdstudio.xyz/api.php?'
            search_urls, page_size, count = [], self.search_size_per_page, 0
            while self.search_size_per_source > count:
                page_rule = copy.deepcopy(default_rule)
                page_rule['count'] = page_size
                page_rule['pages'] = int(count // page_size) + 1
                search_urls.append(base_url + urlencode(page_rule))
                count += page_size
        # return
        return search_urls
    '''_parsegdstudioapi'''
    def _parsegdstudioapi(self, search_results, song_infos: list = [], request_overrides: dict = None):
        # init
        request_overrides = request_overrides or {}
        # parse
        for search_result in search_results:
            # --download results
            if (not isinstance(search_result, dict)) or ('id' not in search_result) or ('raw' not in search_result):
                continue
            song_info = SongInfo(source=self.source)
            for quality in ['play_path_64', 'play_path_aacv164', 'play_path_32', 'play_path_aacv224']:
                download_url: str = search_result['raw'].get(quality, '')
                if not download_url: continue
                song_info = SongInfo(
                    source=self.source, download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                    raw_data={'search': search_result, 'download': {}, 'lyric': {}}, lyric='NULL', duration_s=search_result['raw'].get('duration', 0), 
                    duration=seconds2hms(search_result['raw'].get('duration', 0)), file_size='NULL', ext=download_url.split('?')[0].split('.')[-1],
                    song_name=legalizestring(search_result['raw'].get('title', 'NULL'), replace_null_string='NULL'), 
                    singers=legalizestring(search_result['raw'].get('nickname', 'NULL'), replace_null_string='NULL'), 
                    album=legalizestring(search_result['raw'].get('album_title', 'NULL'), replace_null_string='NULL'),
                    identifier=search_result['id'],
                )
                if song_info.with_valid_download_url: break
            if not song_info.with_valid_download_url: continue
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            ext, file_size = song_info.download_url_status['probe_status']['ext'], song_info.download_url_status['probe_status']['file_size']
            if file_size and file_size != 'NULL': song_info.file_size = file_size
            if ext and ext != 'NULL': song_info.ext = ext
            # --append to song_infos
            song_infos.append(song_info)
            # --judgement for search_size
            if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
        # return
        return song_infos
    '''_parseofficialapi'''
    def _parseofficialapi(self, search_results, song_infos: list = [], request_overrides: dict = None):
        # init
        request_overrides = request_overrides or {}
        for search_result in search_results['data']['track']['docs']:
            # --download results
            if (not isinstance(search_result, dict)) or ('trackUrl' not in search_result):
                continue
            song_info = SongInfo(source=self.source)
            track_id = search_result.get('trackUrl').strip('/').split('/')[-1]
            for quality in [2, 1, 0]:
                params = {"device": "web", "trackId": track_id, "trackQualityLevel": quality}
                try:
                    resp = self.get(f"https://www.ximalaya.com/mobile-playpage/track/v3/baseInfo/{int(time.time() * 1000)}", params=params, **request_overrides)
                    resp.raise_for_status()
                    download_result = resp2json(resp=resp)
                    track_info = safeextractfromdict(download_result, ['trackInfo'], {})
                    if not track_info: continue
                except:
                    continue
                for encrypted_url in sorted(safeextractfromdict(track_info, ['playUrlList'], []), key=lambda x: int(x['fileSize']), reverse=True):
                    if not isinstance(encrypted_url, dict): continue
                    download_url = self._decrypturl(encrypted_url.get('url', ''))
                    if not download_url: continue
                    song_info = SongInfo(
                        source=self.source, download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, lyric='NULL', duration_s=track_info.get('duration', 0), 
                        duration=seconds2hms(track_info.get('duration', 0)), file_size_bytes=encrypted_url.get('fileSize', 0), file_size=byte2mb(encrypted_url.get('fileSize', 0)),
                        ext=download_url.split('?')[0].split('.')[-1], identifier=track_id, song_name=legalizestring(search_result.get('title', 'NULL'), replace_null_string='NULL'), 
                        singers=legalizestring(search_result.get('nickname', 'NULL'), replace_null_string='NULL'), 
                        album=legalizestring(safeextractfromdict(search_result, ['albumInfo', 'title'], ''), replace_null_string='NULL'), 
                    )
                    if song_info.with_valid_download_url: break
                if song_info.with_valid_download_url: break
            if not song_info.with_valid_download_url: continue
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            ext, file_size = song_info.download_url_status['probe_status']['ext'], song_info.download_url_status['probe_status']['file_size']
            if file_size and file_size != 'NULL': song_info.file_size = file_size
            if ext and ext != 'NULL': song_info.ext = ext
            # --append to song_infos
            song_infos.append(song_info)
            # --judgement for search_size
            if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)
            # --parse based on selected API
            parsed_search_url = urlparse(search_url)
            if parsed_search_url.hostname in ['music-api.gdstudio.xyz']:
                self._parsegdstudioapi(search_results, song_infos=song_infos, request_overrides=request_overrides)
            else:
                self._parseofficialapi(search_results, song_infos=song_infos, request_overrides=request_overrides)
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos